import os
import logging
from names_generator import generate_name

real_path = os.path.dirname(os.path.realpath(__file__))
ugle_path, _ = os.path.split(real_path)
    

class CustomFormatter(logging.Formatter):
    COLOR_CODE = {
        'DEBUG': '\033[94m',    # Blue
        'INFO': '\033[92m',     # Green
        'WARNING': '\033[93m',  # Yellow
        'ERROR': '\033[91m',    # Red
        'CRITICAL': '\033[95m'  # Purple
    }
    RESET_CODE = '\033[0m'  # Reset to default color

    def format(self, record):
        # Get the original formatted message
        original_message = super().format(record)

        # Shorten the file path
        record.pathname = os.path.basename(record.pathname)
        original_message = super().format(record)

        # Apply color to the log level
        log_level = record.levelname
        color_code = self.COLOR_CODE.get(log_level, '')
        colored_message = f"{color_code}{original_message}{self.RESET_CODE}"
        
        return colored_message


def create_logger() -> logging.Logger:
    """
    Creates the logging object with logger with custom colour properties

    Returns:
        log (logging.Logger): custom logger
    """
    
    # Create a file handler (logs will be stored in "logs/{exp_name}.log")
    exp_name = generate_name()
    log_path = f'{ugle_path}/logs/'

    if not os.path.exists(log_path):
        os.mkdir(log_path)

    latest_file_index = len(os.listdir(log_path))
    uid = f'{latest_file_index}-{exp_name}'

    # Set up the logger
    logger = logging.getLogger(uid)
    logger.setLevel(logging.INFO)

    # Create a stream handler (console output)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)

    file_handler = logging.FileHandler(filename=log_path + uid)
    file_handler.setLevel(logging.DEBUG)

    # Set the custom formatters to the handlers
    color_formatter = CustomFormatter(fmt="[%(name)s--%(asctime)s] %(message)s", datefmt="%d/%m--%I:%M:%S")
    stream_handler.setFormatter(color_formatter)
    file_handler.setFormatter(color_formatter)

    # Add the handlers to the logger
    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)

    return logger


log = create_logger()



import ugle
import ugle.utils as utils
import ugle.datasets as datasets
from ugle.logger import log
from ugle.hpo import ParamRepeatPruner, StopWhenMaxTrialsHit
import time
from omegaconf import OmegaConf, DictConfig, ListConfig
import numpy as np
import optuna
from optuna import Trial
from optuna.samplers import TPESampler
import time
from tqdm import trange
import copy
import torch
from typing import Dict, List, Optional, Tuple
import optuna
import warnings
from os import makedirs, devnull, get_terminal_size
from memory_profiler import memory_usage
import pickle
from torch_geometric.utils import to_undirected
import types

from optuna.exceptions import ExperimentalWarning
warnings.filterwarnings("ignore", category=ExperimentalWarning)
optuna.logging.set_verbosity(optuna.logging.CRITICAL)


def log_trial_result(trial: Trial, results: dict, valid_metrics: list, multi_objective_study: bool):
    """
        Logs the results for a trial.

        Args:
            trial (Trial): trial object
            results (dict): result dictionary for the trial
            valid_metrics (list): Validation metrics used
            multi_objective_study (boolean): indicating whether under multi-objective study or not

    """
    if not multi_objective_study:
        # log validation results
        trial_value = results[valid_metrics[0]]
        log.info(f'Trial {trial.number} finished. Validation performance: {valid_metrics[0]}={trial_value}')
        # log best trial comparison
        new_best_message = f'New best trial {trial.number}'
        if trial.number > 0:
            if trial_value > trial.study.best_value and trial.study.direction.name == 'MAXIMIZE':
                log.info(new_best_message)
            elif trial_value < trial.study.best_value and trial.study.direction.name == 'MINIMIZE':
                log.info(new_best_message)
            else:
                log.info(
                    f'Trial {trial.number} finished. Best trial is {trial.study.best_trial.number} with {valid_metrics[0]}={trial.study.best_value}')
        else:
            log.info(new_best_message)

    else:
        # log trial results
        right_order_results = [results[k] for k in valid_metrics]
        to_log_trial_values = ''.join(
            f'{metric}={right_order_results[i]}, ' for i, metric in enumerate(valid_metrics))
        log.info(f'Trial {trial.number} finished. Validation performance: {to_log_trial_values.rpartition(",")[0]}')
        # log best trial comparison
        if trial.number > 0:
            # get best values for each metric across the best trials
            best_values, associated_trial = ugle.utils.extract_best_trials_info(trial.study, valid_metrics)
            # compare best values for each trial with new values found
            improved_metrics = []
            for i, metric_result in enumerate(right_order_results):
                if (metric_result > best_values[i] and trial.study.directions[i].name == 'MAXIMIZE') or \
                        (metric_result < best_values[i] and trial.study.directions[i].name == 'MINIMIZE'):
                    best_values[i] = metric_result
                    improved_metrics.append(valid_metrics[i])
                    associated_trial[i] = trial.number

            # log best trial and value for each metric
            if improved_metrics:
                improved_metrics_str = ''.join(f'{metric}, ' for metric in improved_metrics)
                improved_metrics_str = improved_metrics_str[:improved_metrics_str.rfind(',')]
                log.info(f'New best trial for metrics: {improved_metrics_str}')
            else:
                log.info('Trial worse than existing across all metrics')

            best_so_far = ''.join(
                f'trial {associated_trial[i]} ({metric}={best_values[i]}), ' for i, metric in enumerate(valid_metrics))
            best_so_far = best_so_far[:best_so_far.rfind(',')]
            log.info(f'Best results so far: {best_so_far}')

    return


class ugleTrainer:
    def __init__(self, model_name: str, cfg=None):
        super(ugleTrainer, self).__init__()

        # get the required functions from the other debugger
        model_trainer = getattr(getattr(ugle.models, model_name), f"{model_name}_trainer")
        self.preprocess_data = types.MethodType(model_trainer.preprocess_data, self)
        self.training_preprocessing = types.MethodType(model_trainer.training_preprocessing, self)
        self.training_epoch_iter = types.MethodType(model_trainer.training_epoch_iter, self)
        self.test = types.MethodType(model_trainer.test, self)

        # if cfg is none then loads the default
        if not cfg:
            cfg = utils.load_model_config(override_model=model_name)
            cfg.trainer.only_testing = False

        # set the correct device for gpu:id / cpu 
        _device = ugle.utils.set_device(cfg.trainer.gpu)
        if _device == 'cpu':
            self.device_name = _device
        else:
            self.device_name = cfg.trainer.gpu
        self.device = torch.device(_device)

        # set the random seed
        utils.set_random(cfg.args.random_seed)
        if cfg.trainer.show_config:
            log.info(OmegaConf.to_yaml(cfg.args))

        # parses the config options for hpo study
        cfg = ugle.utils.process_study_cfg_parameters(cfg)
        self.cfg = cfg

        makedirs(cfg.trainer.models_path, exist_ok=True)
        makedirs(cfg.trainer.results_path, exist_ok=True)

        self.progress_bar = None
        self.model = None
        self.loss_function = None
        self.scheduler = None
        self.optimizers = None
        self.patience_wait = 0
        self.best_loss = 1e9
        self.current_epoch = 0
     

    def load_database(self, dataset: dict=None):
        """
        Function to load the dataset and set the attributes of the dataset to cfg.

        Args:
            dataset (dict): contains an in memory dataset

        Returns: 
            features (np.ndarray): feature matrix 
            label (np.ndarray): labels for each node
            train_adjacency (np.ndarray): train/validation split of the adjacency matrix
            test_adjacency (np.ndarray): test split of the adjacency matrix
        """
        if dataset:
            self.cfg.dataset = 'In_Memory'
        log.info(f'Loading dataset {self.cfg.dataset}')
        
        # if no dataset has been passed to trainer
        if not dataset:
            if 'synth' not in self.cfg.dataset:
                # loads and splits the dataset
                features, label, train_adjacency, test_adjacency = datasets.load_real_graph_data(
                    self.cfg.dataset,
                    self.cfg.trainer.training_to_testing_split,
                    self.cfg.trainer.split_scheme)
            else:
                _, adj_type, feature_type, n_clusters = self.cfg.dataset.split("_")
                n_clusters = int(n_clusters)
                adjacency, features, label = datasets.create_synth_graph(n_nodes=1000, n_features=500, n_clusters=n_clusters, 
                                                                        adj_type=adj_type, feature_type=feature_type)
                train_adjacency, test_adjacency = datasets.split_adj(adjacency,
                                                                     self.cfg.trainer.training_to_testing_split,
                                                                     self.cfg.trainer.split_scheme)
                
        # if an in-memory dataset has been passed
        else:
            assert 'adjacency' in dataset, "dataset did not contain adjacency info"
            assert 'features' in dataset, "dataset did not contain features info"
            assert 'adjacency' in dataset, "dataset did not contain label info"
            assert np.shape(dataset['features'])[0] == np.shape(dataset['adjacency'])[0], "feature and adjacency node dimension mismatch"
            assert np.shape(dataset['features'])[0] == np.shape(dataset['adjacency'])[0], "feature and adjacency node dimension mismatch"
            assert np.shape(dataset['features'])[0] == np.shape(dataset['label'])[0], "label node dimension mismatch with input data"

            log.info('Converting adjacency matrix connections to unweighted')
            dataset['adjacency'] = np.where(dataset['adjacency'] != 0, 1, dataset['adjacency'])

            log.info('Converting adjacency to undirected')
            dataset['adjacency'] = datasets.to_dense_adj(to_undirected(datasets.to_edge_index(dataset['adjacency']))).numpy().squeeze(0)

            train_adjacency, test_adjacency = datasets.split_adj(dataset['adjacency'],
                                                        self.cfg.trainer.training_to_testing_split,
                                                        self.cfg.trainer.split_scheme)
            
            features = dataset['features']
            label = dataset['label']

        # extract database relevant info
        if not self.cfg.args.n_clusters:
            self.cfg.args.n_clusters = np.unique(label).shape[0]
        self.cfg.args.n_nodes = features.shape[0]
        self.cfg.args.n_features = features.shape[1]

        return features, label, train_adjacency, test_adjacency
    
    def eval(self, dataset: dict=None):
        """
        Function to load database, split adjacency, preprocess dataset correctly for the method
        run a HPO if required, find the best model based on the validation data and then test 
        the best model. 

        Args: 
            dataset (dict): optional dataset dictionary that contains an In-memory dataset

        Returns:
            objective_results (dict): performance dictionary that contains results 
                along with associated best args and the metrics that they were best at

        """

        # loads the database to train on
        features, label, validation_adjacency, test_adjacency = self.load_database(dataset)
            
        # creates store for range of hyperparameters optimised over
        self.cfg.hypersaved_args = copy.deepcopy(self.cfg.args)

        log.debug('splitting dataset into train/validation')
        train_adjacency, validation_adjacency = datasets.split_adj(validation_adjacency, 
                                                                   self.cfg.trainer.train_to_valid_split, 
                                                                   self.cfg.trainer.split_scheme)

        # process data for training
        processed_data = self.preprocess_data(features, train_adjacency)
        processed_valid_data = self.preprocess_data(features, validation_adjacency)

        # if only testing then no need to optimise hyperparameters
        if not self.cfg.trainer.only_testing:
            optuna.logging.disable_default_handler()
            # creates the hpo study
            if self.cfg.trainer.multi_objective_study:
                study = optuna.create_study(study_name=f'{self.cfg.model}_{self.cfg.dataset}',
                                            directions=self.cfg.trainer.optimisation_directions,
                                            sampler=TPESampler(seed=self.cfg.args.random_seed, multivariate=True, group=True))
            else:
                study = optuna.create_study(study_name=f'{self.cfg.model}_{self.cfg.dataset}',
                                            direction=self.cfg.trainer.optimisation_directions[0],
                                            sampler=TPESampler(seed=self.cfg.args.random_seed))
            log.info(f"A new hyperparameter study created: {study.study_name}")
            # create the trial pruners that remove repeat suggested parameters
            study_stop_cb = StopWhenMaxTrialsHit(self.cfg.trainer.n_trials_hyperopt, self.cfg.trainer.max_n_pruned)
            prune_params = ParamRepeatPruner(study)
            # adds the best hps from previous seeds to speed up optimisation
            if self.cfg.trainer.suggest_hps_from_previous_seed:
                for hps in self.cfg.trainer.hps_found_so_far:
                    study.enqueue_trial(hps)
            # perform the hpo
            study.optimize(lambda trial: self.train(trial,
                                                    label,
                                                    processed_data,
                                                    validation_adjacency,
                                                    processed_valid_data,
                                                    prune_params=prune_params),
                                n_trials=self.cfg.trainer.n_trials_hyperopt,
                                callbacks=[study_stop_cb])

            # assigns test parameters found in the study
            if not self.cfg.trainer.multi_objective_study:
                params_to_assign = study.best_trial.params
                if params_to_assign != {}:
                    log.info('Best Hyperparameters found: ')
                    for hp_key, hp_val in params_to_assign.items():
                        log.info(f'{hp_key} : {hp_val}')
                self.cfg = utils.assign_test_params(self.cfg, params_to_assign)
        else:
            params_to_assign = 'default'
            study = None

        processed_test_data = self.preprocess_data(features, test_adjacency)
        if self.cfg.trainer.only_testing:
            validation_results = self.train(None, label, processed_data, validation_adjacency, processed_valid_data)
        elif not self.cfg.trainer.multi_objective_study:
            validation_results = study.best_trial.user_attrs['validation_results']
        # if you aren't optimising for more than one metric or are only evaluating hps 
        if (not self.cfg.trainer.multi_objective_study) or self.cfg.trainer.only_testing:                         
            objective_results = []
            for opt_metric in self.cfg.trainer.valid_metrics:
                log.info(f'Evaluating {opt_metric} model on test split')
                self.model.load_state_dict(torch.load(f"{self.cfg.trainer.models_path}{self.cfg.model}_{self.device_name}_{opt_metric}.pt")['model'])
                self.model.to(self.device)
                results = self.testing_loop(label, test_adjacency, processed_test_data, self.cfg.trainer.test_metrics)
                # log test results
                right_order_results = [results[k] for k in self.cfg.trainer.test_metrics]
                to_log_trial_values = ''.join(f'{metric}={right_order_results[i]}, ' for i, metric in enumerate(self.cfg.trainer.test_metrics))
                log.info(f'Test results optimised for {opt_metric}: {to_log_trial_values.rpartition(",")[0]}')
                objective_results.append({'metrics': opt_metric, 'results': results, 
                                          'args': params_to_assign, "validation_results": validation_results})

        # if you have optimised for more than one metric then there are potentially more than one 
        # metrics that need to be optimised for 
        else:
            # extract best hps found for each metrics
            best_values, associated_trial = ugle.utils.extract_best_trials_info(study, self.cfg.trainer.valid_metrics)
            unique_trials = list(np.unique(np.array(associated_trial)))
            objective_results = []

            for idx, best_trial_id in enumerate(unique_trials):
                # find the metrics for which trial is best at
                best_at_metrics = [metric for i, metric in enumerate(self.cfg.trainer.valid_metrics) if associated_trial[i] == best_trial_id]
                best_hp_params = [trial for trial in study.best_trials if trial.number == best_trial_id][0].params

                # log the best hyperparameters and metrics at which they are best
                best_at_metrics_str = ''.join(f'{metric}, ' for metric in best_at_metrics)
                best_at_metrics_str = best_at_metrics_str[:best_at_metrics_str.rfind(',')]
                test_metrics = ''.join(f'{metric}_' for metric in best_at_metrics)
                test_metrics = test_metrics[:test_metrics.rfind(',')]
                log.info(f'Best hyperparameters for metric(s): {best_at_metrics_str} ')
                if not self.cfg.trainer.only_testing:
                    for hp_key, hp_val in best_hp_params.items():
                        log.info(f'{hp_key} : {hp_val}')
                
                # assign best hyperparameters to config
                self.cfg = utils.assign_test_params(self.cfg, best_hp_params)

                # at this point, the self.train() loop should go have saved models for each validation metric 
                # then go through the best at metrics, and do a test for each of the best models 
                for opt_metric in best_at_metrics:
                    # if epoch is the same for all opt_metrics, there will be an unecessary forward pass but we move since unlikely 
                    log.debug(f'Evaluating {opt_metric} model on test split')
                    self.model.load_state_dict(torch.load(f"{self.cfg.trainer.models_path}{self.cfg.model}_{self.device_name}_{opt_metric}.pt")['model'])
                    self.model.to(self.device)
                    results = self.testing_loop(label, test_adjacency, processed_test_data, self.cfg.trainer.test_metrics)
                    
                    # log test results
                    right_order_results = [results[k] for k in self.cfg.trainer.test_metrics]
                    to_log_trial_values = ''.join(f'{metric}={right_order_results[i]}, ' for i, metric in enumerate(self.cfg.trainer.test_metrics))
                    log.info(f'Test results optimised for {opt_metric}: {to_log_trial_values.rpartition(",")[0]}')
                    objective_results.append({'metrics': opt_metric, 'results': results, 
                                              'args': best_hp_params, "validation_results": study.trials[best_trial_id].user_attrs['validation_results']})
                                            
                # re init the args object for assign_test params
                self.cfg.args = copy.deepcopy(self.cfg.hypersaved_args)

                # this isnt the best code but we move
                if self.cfg.trainer.save_model:
                    log.info(f'Saving best version of model for {best_at_metrics}')
                    torch.save({"model": self.model.state_dict(),
                                "args": best_hp_params},
                               f"{self.cfg.trainer.models_path}{self.cfg.model}_{test_metrics}.pt")

        if study and self.cfg.trainer.save_hpo_study:
            pickle.dump(study, open(f"{self.cfg.trainer.results_path}study_{self.cfg.args.random_seed}_{self.cfg.dataset}_{self.cfg.model}.pkl","wb"))

        return objective_results

    def train(self, trial: Trial, label: np.ndarray, processed_data: tuple, validation_adjacency: np.ndarray, processed_valid_data: tuple, prune_params=None):
        """
        Training function for training a GNN under hpo suggested or specified parameters. Contains functions 
        for switching to active devices, progress bar tracking and validation selection. 

        Args:
            trial (Trial): the trial object from the optuna hpo (can be None) 
            label (np.ndarray): the ground-truth for the dataset
            processed_data (tuple): the processed data necessary for the algorithm specified
            validation_adjacency (np.ndarray): the validation adjacency matrix
            processed_valid_data (tuple): tthe processed validation data necessary for the algorithm specified
            prune_params (object): the object for pruning parameters when suggested twice

        Returns: 
            if trial is None: returns validation results
            else: returns validation performance results as tuple for the trial in hpo
        """


        timings = np.zeros(2)
        start = time.time()
        # configuration of args if hyperparameter optimisation phase
        if trial is not None:
            log.info(f'Launching Trial {trial.number}')
            self.cfg.args = copy.deepcopy(self.cfg.hypersaved_args)
            # this will prune the trial if the args have appeared 
            self.cfg.args = ugle.utils.sample_hyperparameters(trial, self.cfg.args, prune_params)

        # process model creation
        processed_data = self.move_to_activedevice(processed_data)
        self.training_preprocessing(self.cfg.args, processed_data)
        self.model.train()

        # create training loop
        self.progress_bar = trange(self.cfg.args.max_epoch, desc='Training...', leave=True, position=0, bar_format='{l_bar}{bar:15}{r_bar}{bar:-15b}', file=open(devnull, 'w'))
        best_so_far = np.zeros(len((self.cfg.trainer.valid_metrics))) - 0.01
        best_epochs = np.zeros(len((self.cfg.trainer.valid_metrics)), dtype=int)
        if 'conductance' in self.cfg.trainer.valid_metrics:
            best_so_far[self.cfg.trainer.valid_metrics.index('conductance')] = 1.01
        patience_waiting = np.zeros((len(self.cfg.trainer.valid_metrics)), dtype=int)

        for self.current_epoch in self.progress_bar:
            if self.current_epoch % self.cfg.trainer.log_interval == 0:
                if self.current_epoch != 0:
                    if not (self.current_epoch - self.cfg.trainer.log_interval == 0 and self.cfg.trainer.calc_memory):
                        try: 
                            nlength_terminal = get_terminal_size().columns
                        except: 
                           nlength_terminal = 80
                        for _ in range((len_prev_line // nlength_terminal) + 1):
                            utils.remove_last_line() 
                    log.info(str(self.progress_bar))
                    len_prev_line = len(log.name) + 19 + len(str(self.progress_bar))
                else:
                    log.info(str(self.progress_bar))
                    len_prev_line = len(log.name) + 19 + len(str(self.progress_bar))

            timings[0] += time.time() - start
            start = time.time()
            # check if validation time
            if self.current_epoch % self.cfg.trainer.validate_every_nepochs == 0:
                # put in validation mode 
                processed_data = self.move_to_cpudevice(processed_data)

                # compute training iteration
                if self.cfg.trainer.calc_memory and self.current_epoch == 0: 
                    mem_usage, results = memory_usage((self.testing_loop, (label, validation_adjacency, processed_valid_data, self.cfg.trainer.valid_metrics)), 
                                                      interval=.005, retval=True) 
                    log.info(f"Max memory usage by testing_loop(): {max(mem_usage):.2f}MB")
                else:
                    results = self.testing_loop(label, validation_adjacency, processed_valid_data, self.cfg.trainer.valid_metrics)
                
                # put data back into training mode
                processed_data = self.move_to_activedevice(processed_data)

                # record best model state over this train/trial for each metric
                for m, metric in enumerate(self.cfg.trainer.valid_metrics):
                    if (results[metric] > best_so_far[m] and metric != 'conductance') or (results[metric] < best_so_far[m] and metric == 'conductance'):
                        best_so_far[m] = results[metric]
                        best_epochs[m] = self.current_epoch
                        # save model for this metric
                        torch.save({"model": self.model.state_dict(), "args": self.cfg.args},
                                    f"{self.cfg.trainer.models_path}{self.cfg.model}_{self.device_name}_{metric}.pt")
                        patience_waiting[m] = 0
                    else:
                        patience_waiting[m] += 1
            else: 
                patience_waiting += 1
            
            timings[1] += time.time() - start
            start = time.time()

            if self.cfg.trainer.calc_memory and self.current_epoch == 0:
                mem_usage, retvals = memory_usage((self.training_epoch_iter, (self.cfg.args, processed_data)), interval=.005, retval=True) 
                loss, data_returned = retvals
                log.info(f"Max memory usage by training_epoch_iter(): {max(mem_usage):.2f}MB")
            else:
                loss, data_returned = self.training_epoch_iter(self.cfg.args, processed_data)
            if data_returned:
                processed_data = data_returned

            # optimise
            for opt in self.optimizers:
                opt.zero_grad()

            if self.cfg.trainer.calc_memory and self.current_epoch == 0: 
                mem_usage = memory_usage((loss.backward), interval=.005)
                log.info(f"Max memory usage by loss.backward(): {max(mem_usage):.2f}MB")
            else:
                loss.backward()

            for opt in self.optimizers:
                opt.step()
            if self.scheduler:
                self.scheduler.step()

            # update progress bar
            self.progress_bar.set_description(f'Training: m-{self.cfg.model}, d-{self.cfg.dataset}, s-{self.cfg.args.random_seed} -- Loss={loss.item():.4f}', refresh=True)
            # if all metrics hit patience then end
            if patience_waiting.all() >= self.cfg.args.patience:
                log.info(f'Early stopping at epoch {self.current_epoch}!')
                break
        
        len_prev_line = len(log.name) + 19 + len(str(self.progress_bar))
        try: 
            nlength_terminal = get_terminal_size().columns
        except: 
            nlength_terminal = 80
        for _ in range((len_prev_line // nlength_terminal) + 1):
            utils.remove_last_line() 
        log.info(str(self.progress_bar))
        # finished training, record time taken
        timings[0] += time.time() - start
        start = time.time()    
        # compute final validation 
        return_results = dict(zip(self.cfg.trainer.valid_metrics, best_so_far))
        processed_data = self.move_to_cpudevice(processed_data)
        #  then record the result on the validation metrics
        valid_results = {}
        # print best performance per validation set
        for m, metric in enumerate(self.cfg.trainer.valid_metrics):
            log.info(f'Best model by validation {metric} ({best_so_far[m]}) at epoch {best_epochs[m]}')
            if trial is None:
                self.model.load_state_dict(torch.load(f"{self.cfg.trainer.models_path}{self.cfg.model}_{self.device_name}_{metric}.pt")['model'])
                self.model.to(self.device)
                results = self.testing_loop(label, validation_adjacency, processed_valid_data, self.cfg.trainer.test_metrics)
                valid_results[metric] = results
            

        timings[1] += time.time() - start
        start = time.time()
        if self.cfg.trainer.calc_time:
            log.info(f"Time training {round(timings[0], 3)}s")
            log.info(f"Time validating {round(timings[1], 3)}s")

        # log trial result in context of the hpo + return info for optimisation
        if trial is None:
            return valid_results
        else:
            trial.set_user_attr("validation_results", valid_results)
            log_trial_result(trial, return_results, self.cfg.trainer.valid_metrics, self.cfg.trainer.multi_objective_study)
            if not self.cfg.trainer.multi_objective_study:
                return return_results[self.cfg.trainer.valid_metrics[0]]
            else:
                right_order_results = [return_results[k] for k in self.cfg.trainer.valid_metrics]
                return tuple(right_order_results)
            
    def testing_loop(self, label: np.ndarray, adjacency: np.ndarray, processed_data: tuple, eval_metrics: ListConfig):
        """
        Process the testing data through the model to get predictions then evaluates those predictions.

        Args:
            label (np.ndarray): the ground-truth for the dataset
            adjacency (np.ndarray): the testing adjacency matrix - necessary for conductance and modularity
            processed_data (tuple): the processed test data ready for the model
            eval_metrics (ListConfig): a list of metrics to evaluate.

        Returns: 
            results (dict): performance results for each of the metrics
        """
        self.model.eval()
        processed_data = self.move_to_activedevice(processed_data)
        preds = self.test(processed_data)
        processed_data = self.move_to_cpudevice(processed_data)
        results, _ = ugle.process.preds_eval(label, preds, sf=4, adj=adjacency, metrics=eval_metrics)
        self.model.train()
        return results

    def preprocess_data(self, features: np.ndarray, adjacency: np.ndarray) -> tuple:
        log.error('NO PROCESSING STEP IMPLEMENTED FOR MODEL')
        pass

    def test(self, processed_data: tuple) -> np.ndarray:
        log.error('NO TESTING STEP IMPLEMENTED')
        pass

    def training_preprocessing(self, args: DictConfig, processed_data: tuple):
        log.error('NO TRAINING PREPROCESSING PROCEDURE')
        pass

    def training_epoch_iter(self, args: DictConfig, processed_data: tuple):
        log.error('NO TRAINING ITERATION LOOP')
        pass

    def move_to_cpudevice(self, data: torch.Tensor):
        return tuple(databite.to(torch.device("cpu"), non_blocking=True) if torch.is_tensor(databite) else databite for databite in data)

    def move_to_activedevice(self, data: torch.Tensor):
        return tuple(databite.to(self.device, non_blocking=True)  if torch.is_tensor(databite) else databite for databite in data)





import torch
from torch import nn
import torch.nn.functional as F

class mvgrl_Discriminator(nn.Module):
    def __init__(self, n_h):
        super(mvgrl_Discriminator, self).__init__()
        self.f_k = nn.Bilinear(n_h, n_h, 1)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Bilinear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, c1, c2, h1, h2, h3, h4, s_bias1=None, s_bias2=None):
        c_x1 = torch.unsqueeze(c1, 1)
        c_x1 = c_x1.expand_as(h1).contiguous()
        c_x2 = torch.unsqueeze(c2, 1)
        c_x2 = c_x2.expand_as(h2).contiguous()

        # positive
        sc_1 = torch.squeeze(self.f_k(h2, c_x1), 2)
        sc_2 = torch.squeeze(self.f_k(h1, c_x2), 2)

        # negetive
        sc_3 = torch.squeeze(self.f_k(h4, c_x1), 2)
        sc_4 = torch.squeeze(self.f_k(h3, c_x2), 2)

        logits = torch.cat((sc_1, sc_2, sc_3, sc_4), 1)
        return logits

class Discriminator(nn.Module):
    def __init__(self, n_h):
        super(Discriminator, self).__init__()
        self.f_k = nn.Bilinear(n_h, n_h, 1)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Bilinear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, c, h_pl, h_mi, s_bias1=None, s_bias2=None):
        c_x = torch.unsqueeze(c, 1)
        c_x = c_x.expand_as(h_pl)

        sc_1 = torch.squeeze(self.f_k(h_pl, c_x), 2)
        sc_2 = torch.squeeze(self.f_k(h_mi, c_x), 2)

        if s_bias1 is not None:
            sc_1 += s_bias1
        if s_bias2 is not None:
            sc_2 += s_bias2

        logits = torch.cat((sc_1, sc_2), 1)

        return logits

# Applies an average on seq, of shape (batch, nodes, features)
# While taking into account the masking of msk
class AvgReadout(nn.Module):
    def __init__(self):
        super(AvgReadout, self).__init__()

    def forward(self, seq, msk):
        if msk is None:
            return torch.mean(seq, 1)
        else:
            msk = torch.unsqueeze(msk, -1)
            return torch.sum(seq * msk, 1) / torch.sum(msk)

class GATLayer(nn.Module):
    """
    Simple GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(self, in_features, out_features, alpha=0.2):
        super(GATLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha

        self.W = nn.Parameter(torch.zeros(size=(in_features, out_features)))
        nn.init.xavier_uniform_(self.W.data, gain=1.414)

        self.a_self = nn.Parameter(torch.zeros(size=(out_features, 1)))
        nn.init.xavier_uniform_(self.a_self.data, gain=1.414)

        self.a_neighs = nn.Parameter(torch.zeros(size=(out_features, 1)))
        nn.init.xavier_uniform_(self.a_neighs.data, gain=1.414)

        self.leakyrelu = nn.LeakyReLU(self.alpha)

    def forward(self, input, adj, M, concat=True):
        h = torch.mm(input, self.W)

        attn_for_self = torch.mm(h, self.a_self)  # (N,1)
        attn_for_neighs = torch.mm(h, self.a_neighs)  # (N,1)
        attn_dense = attn_for_self + torch.transpose(attn_for_neighs, 0, 1)
        attn_dense = torch.mul(attn_dense, M)
        attn_dense = self.leakyrelu(attn_dense)  # (N,N)

        zero_vec = -9e15 * torch.ones_like(adj)
        adj = torch.where(adj > 0, attn_dense, zero_vec)
        attention = F.softmax(adj, dim=1)
        h_prime = torch.matmul(attention, h)

        if concat:
            return F.elu(h_prime)
        else:
            return h_prime

    def __repr__(self):
        return (
            self.__class__.__name__
            + " ("
            + str(self.in_features)
            + " -> "
            + str(self.out_features)
            + ")"
        )

class GAT(nn.Module):
    def __init__(self, num_features, hidden_size, embedding_size, alpha):
        super(GAT, self).__init__()
        self.hidden_size = hidden_size
        self.embedding_size = embedding_size
        self.alpha = alpha
        self.conv1 = GATLayer(num_features, hidden_size, alpha)
        self.conv2 = GATLayer(hidden_size, embedding_size, alpha)

    def forward(self, x, adj, M):
        h = self.conv1(x, adj, M)
        h = self.conv2(h, adj, M)
        z = F.normalize(h, p=2, dim=1)
        A_pred = self.dot_product_decode(z)
        return A_pred, z

    def dot_product_decode(self, Z):
        A_pred = torch.sigmoid(torch.matmul(Z, Z.t()))
        return A_pred

class GCN(nn.Module):
    def __init__(self, in_ft, out_ft, act, bias=True, skip=False):
        super(GCN, self).__init__()
        self.fc = nn.Linear(in_ft, out_ft, bias=False)

        if act == 'prelu':
            self.act = nn.PReLU()
        elif act == 'selu':
            self.act = nn.SELU()
        else:
            self.act = act

        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(out_ft))
            torch.nn.init.normal_(self.bias, mean=0.0, std=0.1)
        else:
            self.register_parameter('bias', None)

        if skip:
            self.skip = nn.Parameter(torch.FloatTensor(out_ft))
            torch.nn.init.normal_(self.skip, mean=0.0, std=0.1)
        else:
            self.register_parameter('skip', None)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    # Shape of seq: (batch, nodes, features)
    def forward(self, seq, adj, sparse=False, skip=False):
        seq_fts = self.fc(seq)

        if skip:
            skip_out = seq_fts * self.skip

        if sparse:
            out = torch.unsqueeze(torch.spmm(adj, torch.squeeze(seq_fts, 0)), 0)
        else:
            out = torch.bmm(adj, seq_fts)

        if skip:
            out += skip_out

        if self.bias is not None:
            out += self.bias

        return self.act(out)


from ugle.logger import log
import numpy as np
import optuna
from optuna.trial import TrialState
from typing import Dict, List, Optional
from collections import defaultdict
import optuna


class StopWhenMaxTrialsHit:
    def __init__(self, max_n_trials: int, max_n_pruned: int):
        self.max_n_trials = max_n_trials
        self.max_pruned = max_n_pruned
        self.completed_trials = 0
        self.pruned_in_a_row = 0

    def __call__(self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial) -> None:
        if trial.state == optuna.trial.TrialState.COMPLETE:
            self.completed_trials += 1
            self.pruned_in_a_row = 0
        elif trial.state == optuna.trial.TrialState.PRUNED:
            self.pruned_in_a_row += 1

        if self.completed_trials >= self.max_n_trials:
            log.info('Stopping Study for Reaching Max Number of Trials')
            study.stop()

        if self.pruned_in_a_row >= self.max_pruned:
            log.info('Stopping Study for Reaching Max Number of Pruned Trials in a Row')
            study.stop()


class ParamRepeatPruner:
    """Prunes repeated trials, which means trials with the same parameters won't waste time/resources."""

    def __init__(self, study: optuna.study.Study, repeats_max: int = 0, 
                       should_compare_states: List[TrialState] = [TrialState.COMPLETE],
                       compare_unfinished: bool = True,
    ):
        """
        Args:
            study (optuna.study.Study): Study of the trials.
            repeats_max (int, optional): Instead of prunning all of them (not repeating trials at all, repeats_max=0)
                 you can choose to repeat them up to a certain number of times, useful if your optimization function 
                 is not deterministic and gives slightly different results for the same params. Defaults to 0.
            should_compare_states (List[TrialState], optional): By default it only skips the trial if the paremeters are 
                equal to existing COMPLETE trials, so it repeats possible existing FAILed and PRUNED trials. If you also want to 
                skip these trials then use [TrialState.COMPLETE,TrialState.FAIL,TrialState.PRUNED] for example. Defaults to 
                [TrialState.COMPLETE].
            compare_unfinished (bool, optional): Unfinished trials (e.g. `RUNNING`) are treated like COMPLETE ones, 
                if you don't want this behavior change this to False. Defaults to True.
        """
        self.should_compare_states = should_compare_states
        self.repeats_max = repeats_max
        self.repeats: Dict[int, List[int]] = defaultdict(lambda: [], {})
        self.unfinished_repeats: Dict[int, List[int]] = defaultdict(lambda: [], {})
        self.compare_unfinished = compare_unfinished
        self.study = study

    @property
    def study(self) -> Optional[optuna.study.Study]:
        return self._study

    @study.setter
    def study(self, study):
        self._study = study
        if self.study is not None:
            self.register_existing_trials()

    def register_existing_trials(self):
        """In case of studies with existing trials, it counts existing repeats"""
        trials = self.study.trials
        trial_n = len(trials)
        for trial_idx, trial_past in enumerate(self.study.trials[1:]):
            self.check_params(trial_past, False, -trial_n + trial_idx)

    def prune(self):
        self.check_params()

    def should_compare(self, state):
        return any(state == state_comp for state_comp in self.should_compare_states)

    def clean_unfinised_trials(self):
        trials = self.study.trials
        finished = []
        for key, value in self.unfinished_repeats.items():
            if self.should_compare(trials[key].state):
                for t in value:
                    self.repeats[key].append(t)
                finished.append(key)

        for f in finished:
            del self.unfinished_repeats[f]

    def check_params(self, trial: Optional[optuna.trial.BaseTrial] = None, prune_existing=True,
                           ignore_last_trial: Optional[int] = None):
        if self.study is None:
            return
        trials = self.study.trials
        if trial is None:
            trial = trials[-1]
            ignore_last_trial = -1

        self.clean_unfinised_trials()

        self.repeated_idx = -1
        self.repeated_number = -1
        for idx_p, trial_past in enumerate(trials[:ignore_last_trial]):
            should_compare = self.should_compare(trial_past.state)
            should_compare |= (
                self.compare_unfinished and not trial_past.state.is_finished()
            )
            if should_compare and trial.params == trial_past.params:
                if not trial_past.state.is_finished():
                    self.unfinished_repeats[trial_past.number].append(trial.number)
                    continue
                self.repeated_idx = idx_p
                self.repeated_number = trial_past.number
                break

        if self.repeated_number > -1:
            self.repeats[self.repeated_number].append(trial.number)
        if len(self.repeats[self.repeated_number]) > self.repeats_max:
            if prune_existing:
                log.info('Pruning Trial for Suggesting Duplicate Parameters')
                raise optuna.exceptions.TrialPruned()

        return self.repeated_number

    def get_value_of_repeats(self, repeated_number: int, func=lambda value_list: np.mean(value_list)):
        if self.study is None:
            raise ValueError("No study registered.")
        trials = self.study.trials
        values = (
            trials[repeated_number].value,
            *(
                trials[tn].value
                for tn in self.repeats[repeated_number]
                if trials[tn].value is not None
            ),
        )
        return func(values)
    


import numpy as np
import scipy.sparse as sp
import torch
from scipy.linalg import fractional_matrix_power, inv
from sklearn.metrics import f1_score, normalized_mutual_info_score
from scipy.optimize import linear_sum_assignment
import math
from line_profiler import LineProfiler
from typing import Tuple

SMOOTH_K_TOLERANCE = 1e-5
MIN_K_DIST_SCALE = 1e-3
NPY_INFINITY = np.inf


def compute_ppr(adj: np.ndarray, alpha: int = 0.2, self_loop: bool = True) -> float:
    """
    Computes personalized PageRank diffusion.

    Args:
        adj (np.ndarray): Adjacency matrix
        alpha (float): Teleport probability (default: 0.2)
        self_loop (bool): Whether to add self-loops (default: True)

    Returns:
        float: Personalized PageRank matrix
    """
    if self_loop:
        adj = adj + np.eye(adj.shape[0])  # A^ = A + I_n
    d = np.diag(np.sum(adj, 1))  # D^ = Sigma A^_ii
    dinv = fractional_matrix_power(d, -0.5)  # D^(-1/2)
    at = np.matmul(np.matmul(dinv, adj), dinv)  # A~ = D^(-1/2) x A^ x D^(-1/2)
    return alpha * inv((np.eye(adj.shape[0]) - (1 - alpha) * at))  # a(I_n-(1-a)A~)^-1


def modularity(adjacency: np.ndarray, preds: np.ndarray) -> float:
    """
    Computes modularity.

    Args:
        adjacency (np.ndarray): Adjacency matrix
        preds (np.ndarray): Predicted cluster assignments

    Returns:
        float: Modularity score
    """
    adjacency = sp.coo_matrix(adjacency).tocsr()
    degrees = adjacency.sum(axis=0).A1
    m = degrees.sum()
    result = 0
    for cluster_id in np.unique(preds):
        cluster_indices = np.where(preds == cluster_id)[0]
        adj_submatrix = adjacency[cluster_indices, :][:, cluster_indices]
        degrees_submatrix = degrees[cluster_indices]
        result += np.sum(adj_submatrix) - (np.sum(degrees_submatrix) ** 2) / m
    return result / m


def conductance(adjacency: np.ndarray, preds: np.ndarray) -> float:
    """
    Computes conductance.

    Args:
        adjacency (np.ndarray): Adjacency matrix
        preds (np.ndarray): Predicted cluster assignments

    Returns:
        float: Conductance score
    """
    if len(np.unique(preds)) == 1:
        return 1.
    inter = 0
    intra = 0
    cluster_idx = np.zeros(adjacency.shape[0], dtype=bool)
    for cluster_id in np.unique(preds):
        cluster_idx[:] = 0
        cluster_idx[np.where(preds == cluster_id)[0]] = 1
        adj_submatrix = adjacency[cluster_idx, :]
        inter += np.sum(adj_submatrix[:, cluster_idx])
        intra += np.sum(adj_submatrix[:, ~cluster_idx])
    return intra / (inter + intra)


def preds_eval(labels: np.ndarray, preds: np.ndarray, sf=4, adj: np.ndarray = None, 
               metrics=['nmi', 'f1']) -> Tuple[dict, np.ndarray]:
    """
    Evaluates predictions given metrics.

    Args:
        labels (np.ndarray): True labels
        preds (np.ndarray): Predicted labels
        sf (int): Significant figures for rounding (default: 4)
        adj (np.ndarray): Adjacency matrix (default: None)
        metrics (list): List of metrics to compute (default: ['nmi', 'f1'])

    Returns:
        results (dict): Results dictionary 
        eval_preds (np.ndarray): Evaluated predictions
    """
    # returns the correct predictions to match labels
    eval_preds, _ = hungarian_algorithm(labels, preds)
    results = {}

    if 'nmi' in metrics:
        nmi = normalized_mutual_info_score(labels, eval_preds)
        results['nmi'] = float(round(nmi, sf))

    if 'f1' in metrics:
        f1 = f1_score(labels, eval_preds, average='macro')
        results['f1'] = float(round(f1, sf))

    if 'snmi' in metrics:
        true_num_clusters = len(np.unique(labels))
        found_num_clusters = len(np.unique(preds))

        scaling_k = np.exp(-(np.abs(true_num_clusters - found_num_clusters) / true_num_clusters))
        snmi = scaling_k * nmi
        results['snmi'] = float(round(snmi, sf))

    if 'modularity' in metrics:
        assert adj is not None, 'adj not provided'
        results['modularity'] = float(round(modularity(adj, eval_preds), sf))
    if 'conductance' in metrics:
        assert adj is not None, 'adj not provided'
        results['conductance'] = float(round(conductance(adj, eval_preds), sf))

    if 'n_clusters' in metrics:
        results['n_clusters'] = len(np.unique(preds))

    return results, eval_preds


def hungarian_algorithm(labels: np.ndarray, preds: np.ndarray, col_ind=None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Hungarian algorithm for prediction reassignment.

    Args:
        labels (np.ndarray): True labels
        preds (np.ndarray): Predicted labels
        col_ind (np.ndarray): Column indices for reassignment (default: None)

    Returns:
        preds (np.ndarray): Reassigned predictions 
        col_ind (np.ndarray): Column indices for reassigning future predictions
    """
    labels = labels.astype(int)
    assert preds.size == labels.size
    D = max(preds.max(), labels.max()) + 1
    w = np.zeros((D, D), dtype=int)
    for i in range(preds.size):
        w[preds[i], labels[i]] += 1

    if col_ind is None:
        row_ind, col_ind = linear_sum_assignment(w.max() - w)
        preds = col_ind[preds]

    else:
        preds = [col_ind[int(i)] for i in preds]
        preds = np.asarray(preds)

    return preds, col_ind


def preprocess_features(features: np.ndarray):
    """
    Row-normalize feature matrix and convert to tuple representation.

    Args:
        features: Input feature matrix

    Returns:
        np.ndarray: Preprocessed feature matrix
    """
    rowsum = np.array(features.sum(1))
    nonzero_indexes = np.argwhere(rowsum != 0).flatten()
    r_inv = np.zeros_like(rowsum, dtype=float)
    r_inv[nonzero_indexes] = np.power(rowsum[nonzero_indexes], -1)
    r_inv[np.isinf(r_inv)] = 0.
    r_inv[np.isnan(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    return r_mat_inv.dot(features)


def normalize_adj(adj: np.ndarray):
    """
    Symmetrically normalize adjacency matrix.

    Args:
        adj (np.ndarray): Input adjacency matrix

    Returns:
        sp.coo_matrix: Normalized adjacency matrix
    """
    adj = sp.coo_matrix(adj)
    rowsum = np.array(adj.sum(1))
    d_inv_sqrt = np.power(rowsum, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
    d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
    return adj.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt).tocoo()


def sparse_mx_to_torch_sparse_tensor(sparse_mx: sp.csr_matrix) -> torch.FloatTensor:
    """
    Convert a scipy sparse matrix to a torch sparse tensor.

    Args:
        sparse_mx (sp.csr_matrix): Input sparse matrix

    Returns:
        torch.FloatTensor: Torch sparse tensor
    """
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(int))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)


def euclidean_distance(point1: np.ndarray, point2: np.ndarray):
    """
    Calculate the Euclidean distance between two points.

    Args:
        point1 (np.ndarray): First point
        point2 (np.ndarray): Second point

    Returns:
        float: Euclidean distance between the two points
    """
    point_iterable = zip(point1, point2)
    distance = math.sqrt(sum([(y - x) ** 2 for x, y in point_iterable]))
    return distance


import os
from omegaconf import OmegaConf, DictConfig
import numpy as np
import networkx as nx
import zipfile
import gdown
from pathlib import Path
import shutil
import torch
import copy
import random
import scipy.sparse as sp
from typing import Union, Tuple
from ugle.logger import log, ugle_path
from torch_geometric.utils import to_dense_adj, stochastic_blockmodel_graph
import torch
from torch_geometric.transforms import ToUndirected
from torch_geometric.datasets import Coauthor
from torch_geometric.datasets.amazon import Amazon

google_store_datasets = ['acm', 'amac', 'amap', 'bat', 'citeseer', 'cora', 'cocs', 'dblp', 'eat', 'uat', 'pubmed',
                          'texas', 'wisc', 'cornell']
big_datasets = ['Physics', 'CS', 'Photo', 'Computers']
all_datasets = (google_store_datasets + big_datasets)


def check_data_presence(dataset_name: str) -> bool:
    """
    Checks for dataset presence in local memory.

    Args:
        dataset_name (str): Dataset name to check

    Returns:
        bool: True if dataset is present, False otherwise
    """
    dataset_path = ugle_path + f'/data/{dataset_name}'
    if not os.path.exists(dataset_path):
        return False
    elif not os.path.exists(f'{dataset_path}/{dataset_name}_feat.npy'):
        return False
    elif not os.path.exists(f'{dataset_path}/{dataset_name}_label.npy'):
        return False
    elif not os.path.exists(f'{dataset_path}/{dataset_name}_adj.npy'):
        return False
    else:
        return True


def download_graph_data(dataset_name: str) -> bool:
    """
    Downloads a graph dataset.

    Args:
        dataset_name (str): Name of the dataset to download

    Returns:
        bool: True if successful, False otherwise
    """
    log.info(f'Downloading {dataset_name}')
    download_link_path = ugle_path + '/data/download_links.yaml'
    download_links = OmegaConf.load(download_link_path)
    url = download_links[dataset_name]
    dataset_path = ugle_path + f'/data/{dataset_name}'
    if not os.path.exists(dataset_path):
        os.mkdir(dataset_path)

    dataset_zip_path = dataset_path + f'/{dataset_name}.zip'
    gdown.download(url=url, output=dataset_zip_path, quiet=False, fuzzy=True)
    log.info('Finished downloading')

    # extract the zip file
    log.info('Extracting dataset')
    with zipfile.ZipFile(dataset_zip_path, 'r') as zip_ref:
        zip_ref.printdir()
        zip_ref.extractall(dataset_path)
    log.info('Extraction complete')

    # correct the path dir
    extended_path = f'{dataset_path}/{dataset_name}'
    dataset_path += '/'
    if os.path.exists(extended_path):
        log.info('Extraction to wrong location')
        for subdir, dirs, files in os.walk(extended_path):
            for file in files:
                extract_path = os.path.join(subdir, file)
                file = Path(file)
                out_path = os.path.join(dataset_path, file)
                log.info(f'Extracting {extract_path} to ... {out_path} ...')
                shutil.move(Path(extract_path), Path(out_path))

        shutil.rmtree(extended_path)

    return True



def to_edge_index(adjacency: np.ndarray) -> torch.Tensor:
    """
    Converts adjacency in numpy array form to an array of active edges.

    Args:
        adjacency (np.ndarray): Input adjacency matrix

    Returns:
        torch.Tensor: Edge index tensor
    """
    adj_label = sp.coo_matrix(adjacency)
    adj_label = adj_label.todok()

    outwards = [i[0] for i in adj_label.keys()]
    inwards = [i[1] for i in adj_label.keys()]

    adjacency = torch.tensor([outwards, inwards], dtype=int)
    return adjacency


def dropout_edge_undirected(edge_index: torch.Tensor, p: float = 0.5) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Applies dropout to undirected edges in the edge index.

    Args:
        edge_index (torch.Tensor): Edge index tensor
        p (float): Dropout probability (default: 0.5)

    Returns:
        keep_edge_index (torch.Tensor): Tensor for kept edges
        drop_edge_index (torch.Tensor): Tensor for dropped edges
    """
    if p <= 0. or p >= 1.:
        raise ValueError(f'Dropout probability has to be between 0 and 1 -- (got {p})')
    
    edge_index = edge_index[:, torch.where(edge_index[1, :] > edge_index[0, :])[0]]

    row, col = edge_index
    edge_mask = torch.rand(row.size(0)) >= p
    keep_edge_index = edge_index[:, edge_mask]
    drop_edge_index = edge_index[:, torch.ones_like(edge_mask, dtype=bool) ^ edge_mask]

    keep_edge_index = torch.cat([keep_edge_index, keep_edge_index.flip(0)], dim=1)
    drop_edge_index = torch.cat([drop_edge_index, drop_edge_index.flip(0)], dim=1)

    return keep_edge_index, drop_edge_index


def load_real_graph_data(dataset_name: str, test_split: float = 0.5, 
                         split_scheme: str = 'drop_edges') -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Loads the graph dataset and splits the adjacency matrix into two.

    Args:
        dataset_name (str): Name of the dataset
        test_split (float): Percentage of edges to keep for testing (default: 0.5)
        split_scheme (str): Splitting scheme for the adjacency matrix (default: 'drop_edges')
    
    Returns:
        features (np.ndarray): Feature matrix of the graph
        label (np.ndarray): Labelled ground-truth of the graph
        train_adj (np.ndarray): Training adjacency matrix
        test_adj (np.ndarray): Testing adjacency matrix
    
    
    """
    assert dataset_name in all_datasets, f"{dataset_name} not a real dataset"

    if dataset_name in google_store_datasets:
        if not check_data_presence(dataset_name):
            extraction_success = download_graph_data(dataset_name)
            assert extraction_success, f'download/extraction of dataset {dataset_name} failed'
            assert check_data_presence(dataset_name)

        dataset_path = ugle_path + f'/data/{dataset_name}/{dataset_name}'
        features = np.load(dataset_path + "_feat.npy", allow_pickle=True)
        label = np.load(dataset_path + "_label.npy", allow_pickle=True)
        adjacency = np.load(dataset_path + "_adj.npy", allow_pickle=True)

    elif dataset_name in big_datasets:
        dataset_path = ugle_path + f'/data/{dataset_name}'

        if dataset_name in ['Photo', 'Computers']:
            data = Amazon(root=dataset_path, name=dataset_name, transform=ToUndirected(merge=True))[0]
        elif dataset_name in ['CS', 'Physics']:
            data = Coauthor(root=dataset_path, name=dataset_name, transform=ToUndirected(merge=True))[0]
        
        features = data.x.numpy()
        label = data.y.numpy()
        adjacency = to_dense_adj(data.edge_index).numpy().squeeze(0)

    log.debug('Splitting dataset into training/testing')
    train_adj, test_adj = split_adj(adjacency, test_split, split_scheme)

    return features, label, train_adj, test_adj


def compute_datasets_info(dataset_names: list, 
                          compute_stats: bool=False) -> Union[float, float]:
    """
    Computes the information about dataset statistics.

    Args:
        dataset_names (list): List of datasets to analyze
        compute_stats (bool): Whether to compute additional statistics (default: False)

    Returns:
        clustering_x_data (float): Clustering Coefficient
        closeness_y_data (float): Closeness Centralilty 
    """
    clustering_x_data = []
    closeness_y_data = []

    for dataset_name in dataset_names:
        features, label, train_adjacency, test_adjacency = load_real_graph_data(dataset_name, test_split=1.)
        display_string = dataset_name + ' & '  # name
        display_string += str(train_adjacency.shape[0]) + ' & '  # n_nodes
        display_string += str(features.shape[1]) + ' & '  # n_features
        display_string += str(int(np.nonzero(train_adjacency)[0].shape[0])) + ' & '  # n_edges
        display_string += str(len(np.unique(label))) + ' & '  # n_classes
        if compute_stats: 
            nx_g = nx.Graph(train_adjacency)
            clustering = nx.average_clustering(nx_g)
            cercania = nx.closeness_centrality(nx_g)
            cercania = np.mean(list(cercania.values()))

            clustering_x_data.append(clustering)
            closeness_y_data.append(cercania)

            display_string += str(round(clustering, 3)) + ' & '  # clustering coefficient
            display_string += str(round(cercania, 3)) + ' \\\\'  # closeness centrality
        print(display_string)
   
    return clustering_x_data, closeness_y_data


def aug_drop_features(input_feature: Union[np.ndarray, torch.Tensor], 
                      drop_percent: float = 0.2) -> Union[np.ndarray, torch.Tensor]:
    """
    Augmentation by randomly masking features for every node.

    Args:
        input_feature (Union[np.ndarray, torch.Tensor]): Feature matrix
        drop_percent (float): Percent that any feature is dropped (default: 0.2)

    Returns:
        aug_feature (Union[np.ndarray, torch.Tensor]): Augmented feature matrix
    """
    node_num = input_feature.shape[1]
    mask_num = int(node_num * drop_percent)
    node_idx = [i for i in range(node_num)]
    mask_idx = random.sample(node_idx, mask_num)
    aug_feature = input_feature.detach().clone()
    if type(aug_feature) == np.ndarray:
        zeros = np.zeros_like(aug_feature[0][0])
    else:
        zeros = torch.zeros_like(aug_feature[0][0])
    for j in mask_idx:
        aug_feature[0][j] = zeros
    return aug_feature


def aug_drop_adj(input_adj: np.ndarray, drop_percent: float = 0.2) -> Tuple[np.ndarray, np.ndarray]:
    """
    Augmentation by randomly dropping edges with given probability from a dense matrix.

    Args:
        input_adj (np.ndarray): Input adjacency matrix
        drop_percent (float): Percent that any edge is dropped (default: 0.2)

    Returns:
        aug_adj (np.ndarray): Augmented adjacency matrix
        input_adj (np.ndarray): Original adjacency matrix
    """

    edge_index = to_edge_index(input_adj)
    keep_index, drop_index = dropout_edge_undirected(edge_index, p=1-drop_percent)
    aug_adj = to_dense_adj(drop_index).numpy().squeeze(0)

    n_missing = input_adj.shape[0] - aug_adj.shape[0]
    pad_width = ((0, n_missing), (0, n_missing))
    aug_adj = np.pad(aug_adj, pad_width, mode='constant', constant_values=0.)
    assert input_adj.shape == aug_adj.shape

    return aug_adj, input_adj


def numpy_to_edge_index(adjacency: np.ndarray) -> np.ndarray:
    """
    Converts adjacency in numpy array form to an array of active edges.

    Args:
        adjacency (np.ndarray): Input adjacency matrix

    Returns:
        np.ndarray: Edge index array
    """
    adj_label = sp.coo_matrix(adjacency)
    adj_label = adj_label.todok()

    outwards = [i[0] for i in adj_label.keys()]
    inwards = [i[1] for i in adj_label.keys()]

    adjacency = np.array([outwards, inwards], dtype=int)
    return adjacency


class Augmentations:
    """
    A utility for graph data augmentation
    """

    def __init__(self, method='gdc'):
        methods = {"split", "standardize", "katz"}
        assert method in methods
        self.method = method

    @staticmethod
    def _split(features, permute=True):
        """
        Data augmentation is build by spliting data along the feature dimension.

        :param data: the data object to be augmented
        :param permute: Whether to permute along the feature dimension

        """
        perm = np.random.permutation(features.shape[1]) if permute else np.arange(features.shape[1])
        features = features[:, perm]
        size = features.shape[1] // 2
        x1 = features[:, :size]
        x2 = features[:, size:]

        return x1, x2

    @staticmethod
    def _standardize(features):
        """
        Applies a zscore node feature data augmentation.

        :param data: The data to be augmented
        :return: a new augmented instance of the input data
        """
        mean, std = features.mean(), features.std()
        new_data = (features - mean) / (std + 10e-7)
        return new_data

    @staticmethod
    def _katz(features, adjacency, beta=0.1, threshold=0.0001):
        """
        Applies a Katz-index graph topology augmentation

        :param data: The data to be augmented
        :return: a new augmented instance of the input data
        """
        n_nodes = features.shape[0]

        a_hat = adjacency + sp.eye(n_nodes)
        d_hat = sp.diags(
            np.array(1 / np.sqrt(a_hat.sum(axis=1))).reshape(n_nodes))
        a_hat = d_hat @ a_hat @ d_hat
        temp = sp.eye(n_nodes) - beta * a_hat
        h_katz = (sp.linalg.inv(temp.tocsc()) * beta * a_hat).toarray()
        mask = (h_katz < threshold)
        h_katz[mask] = 0.
        edge_index = np.array(h_katz.nonzero())
        edge_attr = torch.tensor(h_katz[h_katz.nonzero()], dtype=torch.float32)

        return edge_index

    def __call__(self, features, adjacency):
        """
        Applies different data augmentation techniques
        """

        if self.method == "katz":
            aug_adjacency = self._katz(features, adjacency)
            aug_adjacency = np.array([aug_adjacency[0], aug_adjacency[1]], dtype=int)
            adjacency = adjacency.todense()
            adjacency = numpy_to_edge_index(adjacency)
            aug_features = features.copy()
        elif self.method == 'split':
            features, aug_features = self._split(features)
            adjacency = adjacency.todense()
            adjacency = numpy_to_edge_index(adjacency)
            aug_adjacency = adjacency.copy()
        elif self.method == "standardize":
            aug_features = self._standardize(features)
            adjacency = adjacency.todense()
            adjacency = numpy_to_edge_index(adjacency)
            aug_adjacency = adjacency.copy()

        return features, adjacency, aug_features, aug_adjacency

    def __str__(self):
        return self.method.title()


def split(n: int, k: int) -> list:
    """
    From a number of nodes and number of clusters, splits the nodes so that each cluster has an interger
    number of nodes.

    Args: 
        n (int): number of nodes
        k (int): number of clusters

    Returns: 
        list: cluster sizes for each cluster
    """
    d, r = divmod(n, k)
    return [d + 1] * r + [d] * (k - r)


def create_synth_graph(n_nodes: int, n_features: int , n_clusters: int, adj_type: str, 
                       feature_type: str = 'random') -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Creates a synthetic graph with specified properties.

    Args:
        n_nodes (int): Number of nodes in the graph
        n_features (int): Number of features for each node
        n_clusters (int): Number of clusters in the graph
        adj_type (str): Type of adjacency matrix ('disjoint', 'random', 'complete')
        feature_type (str): Type of node features ('random', 'complete', 'disjoint') (default: 'random')

    Returns:
        adj (np.ndarray): generated adjacency matrix
        features (np.ndarray): node features
        labels (np.ndarray): ground-truth
    """
    if adj_type == 'disjoint':
        probs = (np.identity(n_clusters)).tolist()
    elif adj_type == 'random':
        probs = (np.ones((n_clusters, n_clusters))/n_clusters).tolist()
    elif adj_type == 'complete':
        probs = np.ones((n_clusters, n_clusters)).tolist()

    cluster_sizes = split(n_nodes, n_clusters)
    adj = to_dense_adj(stochastic_blockmodel_graph(cluster_sizes, probs)).squeeze(0).numpy()
    
    if feature_type == 'random':
        features = torch.normal(mean=0, std=1, size=(n_nodes, n_features)).numpy()
        features = np.where(features > 0., 1, 0)
    elif feature_type == 'complete': 
        features = np.ones((n_nodes, n_features))
    elif feature_type == 'disjoint':
        features = np.zeros((n_nodes, n_features))
        feature_dims_fo_cluster = split(n_features, n_clusters)
        start_feat = 0
        end_feat = feature_dims_fo_cluster[0]
        start_clus = 0
        end_clus = cluster_sizes[0]
        for i in range(len(feature_dims_fo_cluster)):
            features[start_clus:end_clus, start_feat:end_feat] = np.ones_like(features[start_clus:end_clus, start_feat:end_feat])
            if i == len(feature_dims_fo_cluster) - 1:
                break
            start_feat += feature_dims_fo_cluster[i]
            end_feat += feature_dims_fo_cluster[i+1]
            start_clus += cluster_sizes[i]
            end_clus += cluster_sizes[i+1]

    labels = []
    for i in range(n_clusters):
        labels.extend([i] * cluster_sizes[i])
    labels = np.array(labels)

    return adj, features.astype(float), labels


def split_adj(adj: np.ndarray, percent: float, split_scheme: str):
    """
    Splits the adjacency matrix based on the specified split scheme and percentage.

    Args:
        adj (np.ndarray): Input adjacency matrix
        percent (float): Percentage of edges to keep
        split_scheme (str): Splitting scheme ('drop_edges', 'split_edges', 'all_edges', 'no_edges')

    Returns:
        train_adjacency (np.ndarray): Training adjacency matrix
        validation_adjacency (np.ndarray): Validation adjacency matrix
    """
    if split_scheme == 'drop_edges':
        # drops edges from dataset to form new adj 
        if percent != 1.:
            train_adjacency, validation_adjacency = aug_drop_adj(adj, drop_percent=1 - percent)
        else:
            train_adjacency = adj
            validation_adjacency = adj.copy()
    elif split_scheme == 'split_edges':
        # splits the adj via the edges so that no edges in both 
        if percent != 1.:
            train_adjacency, validation_adjacency = aug_drop_adj(adj, drop_percent=1 - percent)
        else:
            train_adjacency = adj
            validation_adjacency = adj.copy()

    elif split_scheme == 'all_edges':
        # makes the adj fully connected 
        train_adjacency = np.ones_like(adj)
        validation_adjacency = np.ones_like(adj)

    elif split_scheme == 'no_edges':
        # makes the adj completely unconnected 
        train_adjacency = np.zeros_like(adj)
        validation_adjacency = np.zeros_like(adj)
    
    return train_adjacency, validation_adjacency


# code insipred from https://github.com/Tiger101010/DAEGC
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.optim import Adam
from fast_pytorch_kmeans import KMeans
import scipy.sparse as sp
import ugle
from ugle.logger import log
from ugle.trainer import ugleTrainer
from ugle.gnn_architecture import GAT

class DAEGC(nn.Module):
    def __init__(self, num_features, hidden_size, embedding_size, alpha, num_clusters, v=1):
        super(DAEGC, self).__init__()
        self.num_clusters = num_clusters
        self.v = v

        # get pretrain model
        self.gat = GAT(num_features, hidden_size, embedding_size, alpha)
        #self.gat.load_state_dict(torch.load(args.pretrain_path, map_location='cpu'))

        # cluster layer
        self.cluster_layer = Parameter(torch.Tensor(num_clusters, embedding_size))
        torch.nn.init.xavier_normal_(self.cluster_layer.data)

    def forward(self, x, adj, M):
        A_pred, z = self.gat(x, adj, M)
        q = self.get_Q(z)

        return A_pred, z, q

    def get_Q(self, z):
        q = 1.0 / (1.0 + torch.sum(torch.pow(z.unsqueeze(1) - self.cluster_layer, 2), 2) / self.v)
        q = q.pow((self.v + 1.0) / 2.0)
        q = (q.t() / torch.sum(q, 1)).t()
        return q


def target_distribution(q):
    weight = q ** 2 / q.sum(0)
    return (weight.t() / weight.sum(1)).t()

def get_M(adj):
    adj_numpy = adj.todense()
    # t_order
    t=2
    row_l1_norms = np.linalg.norm(adj_numpy, ord=1, axis=1)
    tran_prob = adj_numpy / row_l1_norms[:, None]
    M_numpy = sum([np.linalg.matrix_power(tran_prob, i) for i in range(1, t + 1)]) / t
    return torch.Tensor(M_numpy)


class daegc_trainer(ugleTrainer):

    def preprocess_data(self, features, adjacency):
        adjacency = adjacency + sp.eye(adjacency.shape[0])
        adj_label = adjacency.copy()
        adjacency = ugle.process.normalize_adj(adjacency)
        M = get_M(adjacency)
        adj = torch.FloatTensor(adjacency.todense())
        adj_label = torch.FloatTensor(adj_label)
        features = torch.FloatTensor(features)
        features[features != 0.] = 1.
        return features, adj, adj_label, M

    def training_preprocessing(self, args, processed_data):
        features, adj, adj_label, M = processed_data

        log.debug('creating model')
        self.model = DAEGC(num_features=args.n_features, hidden_size=args.hidden_size,
                      embedding_size=args.embedding_size, alpha=args.alpha, num_clusters=args.n_clusters).to(
            self.device)
        optimizer = Adam(self.model.gat.parameters(), lr=args.pre_lr, weight_decay=args.weight_decay)

        log.debug('pretraining')
        best_nmi = 0.
        for pre_epoch in range(args.pre_epoch):
            # training pass
            self.model.train()
            A_pred, z = self.model.gat(features, adj, M)
            loss = F.binary_cross_entropy(A_pred.view(-1), adj_label.view(-1))
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        log.debug('kmeans init estimate')
        with torch.no_grad():
            _, z = self.model.gat(features, adj, M)

        kmeans = kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
        _ = kmeans.fit_predict(z).cpu().numpy()
        
        self.model.cluster_layer.data = kmeans.centroids.clone().detach().to(self.device)

        log.debug('model training')
        optimizer = Adam(self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)

        self.optimizers = [optimizer]
        return


    def training_epoch_iter(self, args, processed_data):
        if len(processed_data) == 4:
            features, adj, adj_label, M = processed_data
        else:
            features, adj, adj_label, M, Q = processed_data


        if self.current_epoch % args.update_interval == 0:
            # update_interval
            A_pred, z, Q = self.model(features, adj, M)
            q = Q.detach().data.cpu().numpy().argmax(1)

        A_pred, z, q = self.model(features, adj, M)
        p = target_distribution(Q.detach())

        kl_loss = F.kl_div(q.log(), p, reduction='batchmean')
        re_loss = F.binary_cross_entropy(A_pred.view(-1), adj_label.view(-1))

        loss = (args.kl_loss_const * kl_loss) + re_loss

        processed_data = (features, adj, adj_label, M, Q)

        return loss, processed_data

    def test(self, processed_data):

        with torch.no_grad():
            features, adj, adj_label, M = processed_data
            _, z, Q = self.model(features, adj, M)
            preds = Q.detach().data.cpu().numpy().argmax(1)

        return preds


# https://github.com/qcydm/VGAER/tree/main/VGAER_codes
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from fast_pytorch_kmeans import KMeans
import math
from ugle.trainer import ugleTrainer
from copy import deepcopy

class GraphConvolution(nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True, act="tanh"):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if act == "tanh":
            self.act = nn.Tanh()
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, adj, input):
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)

        if self.bias is not None:
            output += self.bias

        return self.act(output)

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'


class InnerProductDecoder(nn.Module):
    """Decoder for using inner product for prediction."""
    def __init__(self, dropout, act=torch.sigmoid):
        super(InnerProductDecoder, self).__init__()
        self.dropout = dropout
        self.act = act

    def forward(self, z):
        z = F.dropout(z, self.dropout, training=self.training)
        B_hat = z @ z.t()
        B_hat = F.sigmoid(B_hat)
        return B_hat


class VGAERModel(nn.Module):
    def __init__(self, in_dim, hidden1_dim, hidden2_dim, device):
        super(VGAERModel, self).__init__()
        self.in_dim = in_dim
        self.hidden1_dim = hidden1_dim
        self.hidden2_dim = hidden2_dim

        layers = [GraphConvolution(self.in_dim, self.hidden1_dim, act="tanh"),
                  GraphConvolution(self.hidden1_dim, self.hidden2_dim, act="tanh"),
                  GraphConvolution(self.hidden1_dim, self.hidden2_dim, act="tanh")]
        self.layers = nn.ModuleList(layers)
        self.device = device

    def encoder(self, a_hat, features):

        h = self.layers[0](a_hat, features)
        self.mean = self.layers[1](a_hat, h)
        self.log_std = self.layers[2](a_hat, h)
        gaussian_noise = torch.randn(features.size(0), self.hidden2_dim).to(self.device)
        sampled_z = self.mean + gaussian_noise * torch.exp(self.log_std).to(self.device)
        return sampled_z

    def decoder(self, z):
        adj_rec = torch.sigmoid(torch.matmul(z, z.t()))
        return adj_rec

    def forward(self, a_hat, features):
        z = self.encoder(a_hat, features)
        adj_rec = self.decoder(z)
        return adj_rec, z


def compute_loss_para(adj):
    try: 
        pos_weight = ((adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum())
        norm = adj.shape[0] * adj.shape[0] / float((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)
        weight_mask = adj.view(-1) == 1
        weight_tensor = torch.ones(weight_mask.size(0))
        weight_tensor[weight_mask] = pos_weight
    except:
        weight_tensor = torch.ones_like(adj.view(-1))
        norm = 1.
    return weight_tensor, norm


class vgaer_trainer(ugleTrainer):
    def preprocess_data(self, features, adjacency):
        A = torch.FloatTensor(adjacency)
        A[A != 0] = 1
        A_orig_ten = A.detach().clone()

        # compute B matrix
        K = 1 / (A.sum().item()) * (A.sum(dim=1).reshape(A.shape[0], 1) @ A.sum(dim=1).reshape(1, A.shape[0]))
        feats = A - K

        # compute A_hat matrix
        A = A + torch.eye(A.shape[0])
        D = torch.diag(torch.pow(A.sum(dim=1), -0.5))  # D = D^-1/2
        A_hat = D @ A @ D

        weight_tensor, norm = compute_loss_para(A)
        weight_tensor = weight_tensor
        A_hat = A_hat
        feats = feats

        return A_orig_ten, A_hat, feats, weight_tensor, norm


    def training_preprocessing(self, args, processed_data):

        self.model = VGAERModel(args.n_nodes, args.hidden1, args.hidden2, self.device).to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)

        self.optimizers = [optimizer]

        return

    def training_epoch_iter(self, args, processed_data):
        A_orig_ten, A_hat, feats, weight_tensor, norm = processed_data
        recovered = self.model.forward(A_hat, feats)
        logits = recovered[0]
        hidemb = recovered[1]

        logits = logits.clamp(min=0., max=1.)

        loss = norm * F.binary_cross_entropy(logits.view(-1), A_orig_ten.view(-1), weight=weight_tensor)
        kl_divergence = 0.5 / logits.size(0) * (
                1 + 2 * self.model.log_std - self.model.mean ** 2 - torch.exp(self.model.log_std) ** 2).sum(
            1).mean()
        loss -= kl_divergence

        return loss, None

    def test(self, processed_data):
        A_orig_ten, A_hat, feats, weight_tensor, norm = processed_data

        self.model.eval()
        recovered = self.model.forward(A_hat, feats)
        emb = recovered[1]
        emb = emb.float().clamp(torch.finfo(torch.float32).min, torch.finfo(torch.float32).max)

        kmeans = kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
        preds = kmeans.fit_predict(emb).cpu().numpy()
      
        return preds


# https://github.com/google-research/google-research/blob/master/graph_embedding/dmon/dmon.py
import ugle
import scipy.sparse as sp
from ugle.trainer import ugleTrainer
import torch.nn as nn
import torch
from ugle.gnn_architecture import GCN
import math
from collections import OrderedDict

class DMoN(nn.Module):
    def __init__(self,
                 args,
                 act='selu',
                 do_unpooling=False):
        """Initializes the layer with specified parameters."""
        super(DMoN, self).__init__()
        self.args = args
        self.n_clusters = args.n_clusters
        self.orthogonality_regularization = args.orthogonality_regularization
        self.cluster_size_regularization = args.cluster_size_regularization
        self.dropout_rate = args.dropout_rate
        self.do_unpooling = do_unpooling
        self.gcn = GCN(args.n_features, args.architecture, act=act, skip=True)
        self.transform = nn.Sequential(OrderedDict([
            ('layer1', nn.Linear(args.architecture, args.n_clusters)),
            ('dropout', nn.Dropout(args.dropout_rate)),
        ]))

        def init_weights(m):
            if isinstance(m, nn.Linear):
                nn.init.orthogonal_(m.weight.data, gain=math.sqrt(2))
                if m.bias is not None:
                    m.bias.data.fill_(0.0)

        self.transform.apply(init_weights)

        return


    def forward(self, graph, graph_normalised, features, extra_loss=False):

        gcn_out = self.gcn(features, graph_normalised, sparse=True, skip=True)

        assignments = self.transform(gcn_out).squeeze(0)
        assignments = nn.functional.softmax(assignments, dim=1)

        n_edges = graph._nnz()
        degrees = torch.sparse.sum(graph, dim=0)._values().unsqueeze(1)
        graph_pooled = torch.spmm(torch.spmm(graph, assignments).T, assignments)
        normalizer_left = torch.spmm(assignments.T, degrees)
        normalizer_right = torch.spmm(assignments.T, degrees).T
        normalizer = torch.spmm(normalizer_left, normalizer_right) / 2 / n_edges
        spectral_loss = - torch.trace(graph_pooled - normalizer) / 2 / n_edges
        loss = spectral_loss

        if extra_loss:
            pairwise = torch.spmm(assignments.T, assignments)
            identity = torch.eye(self.n_clusters).to(pairwise.device)
            orthogonality_loss = torch.norm(pairwise / torch.norm(pairwise) -
                                         identity / math.sqrt(float(self.n_clusters)))
            orthogonality_loss *= self.orthogonality_regularization
            loss += orthogonality_loss

            cluster_loss = torch.norm(torch.sum(pairwise, dim=1)) / self.args.n_nodes * math.sqrt(float(self.n_clusters)) - 1
            cluster_loss *= self.cluster_size_regularization
            loss += cluster_loss

        else:
            cluster_sizes = torch.sum(assignments, dim=0)
            cluster_loss = torch.norm(cluster_sizes) / self.args.n_nodes * math.sqrt(float(self.n_clusters)) - 1
            cluster_loss *= self.cluster_size_regularization
            loss += cluster_loss

        return loss

    def embed(self, graph, graph_normalised, features):
        gcn_out = self.gcn(features, graph_normalised, sparse=True)
        assignments = self.transform(gcn_out).squeeze(0)
        assignments = nn.functional.softmax(assignments, dim=1)

        return assignments


class dmon_trainer(ugleTrainer):

    def preprocess_data(self, features, adjacency):
        features = torch.FloatTensor(features)
        features[features != 0.] = 1.


        adjacency = adjacency + sp.eye(adjacency.shape[0])
        adj_label = adjacency.copy()
        adjacency = ugle.process.normalize_adj(adjacency)

        graph_normalised = ugle.process.sparse_mx_to_torch_sparse_tensor(adjacency)

        adj_label = sp.coo_matrix(adj_label)
        graph = ugle.process.sparse_mx_to_torch_sparse_tensor(adj_label)

        return graph, graph_normalised, features

    def training_preprocessing(self, args, processed_data):

        self.model = DMoN(args).to(self.device)
        optimiser = torch.optim.Adam(self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)
        self.optimizers = [optimiser]

        return

    def training_epoch_iter(self, args, processed_data):
        graph, graph_normalised, features = processed_data
        loss = self.model(graph, graph_normalised, features)
        
        return loss, None

    def test(self, processed_data):
        graph, graph_normalised, features = processed_data
        with torch.no_grad():
            assignments = self.model.embed(graph, graph_normalised, features)
            preds = assignments.detach().cpu().numpy().argmax(axis=1)

        return preds







# inspired by https://github.com/PetarV-/DGI
import numpy as np
import scipy.sparse as sp
import torch
import torch.nn as nn
from fast_pytorch_kmeans import KMeans
import ugle
from ugle.trainer import ugleTrainer
from ugle.gnn_architecture import GCN, AvgReadout, Discriminator

class DGI(nn.Module):
    def __init__(self, n_in, n_h, activation):
        super(DGI, self).__init__()
        self.gcn = GCN(n_in, n_h, activation)
        self.read = AvgReadout()

        self.sigm = nn.Sigmoid()

        self.disc = Discriminator(n_h)

    def forward(self, seq1, seq2, adj, msk, samp_bias1, samp_bias2):
        h_1 = self.gcn(seq1, adj, sparse=True)

        c = self.read(h_1, msk)
        c = self.sigm(c)

        h_2 = self.gcn(seq2, adj, sparse=True)

        ret = self.disc(c, h_1, h_2, samp_bias1, samp_bias2)

        return ret

    # Detach the return variables
    def embed(self, seq, adj, msk):
        h_1 = self.gcn(seq, adj, sparse=True)
        c = self.read(h_1, msk)

        return h_1.detach(), c.detach()


class dgi_trainer(ugleTrainer):

    def preprocess_data(self, features, adjacency):
        adjacency = adjacency + sp.eye(adjacency.shape[0])
        adjacency = ugle.process.normalize_adj(adjacency)
        adj = ugle.process.sparse_mx_to_torch_sparse_tensor(adjacency)
        features = ugle.process.preprocess_features(features)
        features = torch.FloatTensor(features[np.newaxis])

        return adj, features

    def training_preprocessing(self, args, processed_data):
        self.model = DGI(args.n_features, args.hid_units, args.nonlinearity).to(self.device)
        optimiser = torch.optim.Adam(self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)
        self.loss_function = nn.BCEWithLogitsLoss()
        self.optimizers = [optimiser]

        return

    def training_epoch_iter(self, args, processed_data):
        adj, features = processed_data
        idx = np.random.permutation(args.n_nodes)
        shuf_fts = features[:, idx, :]
        lbl_1 = torch.ones(args.batch_size, args.n_nodes)
        lbl_2 = torch.zeros(args.batch_size, args.n_nodes)
        lbl = torch.cat((lbl_1, lbl_2), 1)
        shuf_fts = shuf_fts.to(self.device)
        lbl = lbl.to(self.device)
        logits = self.model(features, shuf_fts, adj, None, None, None)
        loss = self.loss_function(logits, lbl)

        return loss, None

    def test(self, processed_data):
        adj, features = processed_data
        with torch.no_grad():
            embeds, _ = self.model.embed(features, adj, None)
            embeds = embeds.squeeze(0)

        kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
        preds = kmeans.fit_predict(embeds).cpu().numpy()

        return preds



"""
ugle.

This is a repository for investigating implementations of GNNs for unsupervised clustering.
"""
__version__ = "0.6.0"
__author__ = 'William Leeney'

from pkgutil import walk_packages
from importlib import import_module

def import_submodules(package: str, recursive: bool=True):
    """ Import all submodules of a module, recursively, including subpackages

    Args:
        package (str): package (name or actual module)
    Returns:
        results dict[str, ModuleType]: all submodules
    """
    if isinstance(package, str):
        package = import_module(package)
    results = {}
    for loader, name, is_pkg in walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        results[full_name] = import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results

import_submodules(__name__)



# https://github.com/CRIPAC-DIG/GRACE
import torch
import torch.nn as nn
import torch.nn.functional as F
import ugle
import scipy.sparse as sp
import numpy as np
from fast_pytorch_kmeans import KMeans
import warnings
warnings.filterwarnings('ignore')
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=DeprecationWarning)
    from torch_geometric.utils import dropout_adj

from torch_geometric.nn import GCNConv
from ugle.trainer import ugleTrainer


class Encoder(torch.nn.Module):
    def __init__(self, in_channels: int, out_channels: int, activation,
                 base_model, k: int = 2):
        super(Encoder, self).__init__()
        self.base_model = base_model

        assert k >= 2
        self.k = k
        self.conv = [base_model(in_channels, 2 * out_channels)]
        for _ in range(1, k-1):
            self.conv.append(base_model(2 * out_channels, 2 * out_channels))
        self.conv.append(base_model(2 * out_channels, out_channels))
        self.conv = nn.ModuleList(self.conv)

        self.activation = activation

    def forward(self, x: torch.Tensor, edge_index: torch.Tensor):
        for i in range(self.k):
            x = self.activation(self.conv[i](x, edge_index))
        return x


class Model(torch.nn.Module):
    def __init__(self, encoder: Encoder, num_hidden: int, num_proj_hidden: int,
                 tau: float = 0.5):
        super(Model, self).__init__()
        self.encoder: Encoder = encoder
        self.tau: float = tau

        self.fc1 = torch.nn.Linear(num_hidden, num_proj_hidden)
        self.fc2 = torch.nn.Linear(num_proj_hidden, num_hidden)

    def forward(self, x: torch.Tensor,
                edge_index: torch.Tensor) -> torch.Tensor:
        return self.encoder(x, edge_index)

    def projection(self, z: torch.Tensor) -> torch.Tensor:
        z = F.elu(self.fc1(z))
        return self.fc2(z)

    def sim(self, z1: torch.Tensor, z2: torch.Tensor):
        z1 = F.normalize(z1)
        z2 = F.normalize(z2)
        return torch.mm(z1, z2.t())

    def semi_loss(self, z1: torch.Tensor, z2: torch.Tensor):
        f = lambda x: torch.exp(x / self.tau)
        refl_sim = f(self.sim(z1, z1))
        between_sim = f(self.sim(z1, z2))

        return -torch.log(
            between_sim.diag()
            / (refl_sim.sum(1) + between_sim.sum(1) - refl_sim.diag()))

    def batched_semi_loss(self, z1: torch.Tensor, z2: torch.Tensor,
                          batch_size: int):
        # Space complexity: O(BN) (semi_loss: O(N^2))
        device = z1.device
        num_nodes = z1.size(0)
        num_batches = (num_nodes - 1) // batch_size + 1
        f = lambda x: torch.exp(x / self.tau)
        indices = torch.arange(0, num_nodes).to(device)
        losses = []

        for i in range(num_batches):
            mask = indices[i * batch_size:(i + 1) * batch_size]
            refl_sim = f(self.sim(z1[mask], z1))  # [B, N]
            between_sim = f(self.sim(z1[mask], z2))  # [B, N]

            losses.append(-torch.log(
                between_sim[:, i * batch_size:(i + 1) * batch_size].diag()
                / (refl_sim.sum(1) + between_sim.sum(1)
                   - refl_sim[:, i * batch_size:(i + 1) * batch_size].diag())))

        return torch.cat(losses)

    def loss(self, z1: torch.Tensor, z2: torch.Tensor,
             mean: bool = True, batch_size: int = 0):
        h1 = self.projection(z1)
        h2 = self.projection(z2)

        if batch_size == 0:
            l1 = self.semi_loss(h1, h2)
            l2 = self.semi_loss(h2, h1)
        else:
            l1 = self.batched_semi_loss(h1, h2, batch_size)
            l2 = self.batched_semi_loss(h2, h1, batch_size)

        ret = (l1 + l2) * 0.5
        ret = ret.mean() if mean else ret.sum()

        return ret


def drop_feature(x, drop_prob):
    drop_mask = torch.empty(
        (x.size(1), ),
        dtype=torch.float32,
        device=x.device).uniform_(0, 1) < drop_prob
    x = x.clone()
    x[:, drop_mask] = 0

    return x

class grace_trainer(ugleTrainer):
    def preprocess_data(self, features, adjacency):
        adj_label = sp.coo_matrix(adjacency)
        adj_label = adj_label.todok()

        outwards = [i[0] for i in adj_label.keys()]
        inwards = [i[1] for i in adj_label.keys()]

        adj = torch.from_numpy(np.array([outwards, inwards], dtype=int))

        data = torch.FloatTensor(features)

        return data, adj

    def training_preprocessing(self, args, processed_data):

        activation = ({'relu': F.relu, 'prelu': nn.PReLU()})[args.activation]
        base_model = ({'GCNConv': GCNConv})[args.base_model]

        encoder = Encoder(args.n_features, args.num_hidden, activation,
                          base_model=base_model, k=args.num_layers).to(self.device)
        self.model = Model(encoder, args.num_hidden, args.num_proj_hidden, args.tau).to(self.device)
        optimizer = torch.optim.Adam(
            self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)

        self.optimizers = [optimizer]

        return

    def training_epoch_iter(self, args, processed_data):
        data, adj = processed_data

        adj_1 = dropout_adj(adj, p=args.drop_edge_rate_1, force_undirected=True)[0]
        adj_2 = dropout_adj(adj, p=args.drop_edge_rate_2, force_undirected=True)[0]

        x_1 = ugle.datasets.aug_drop_features(data, drop_percent=args.drop_feature_rate_1)
        x_2 = ugle.datasets.aug_drop_features(data, drop_percent=args.drop_feature_rate_2)

        z1 = self.model(x_1, adj_1)
        z2 = self.model(x_2, adj_2)

        loss = self.model.loss(z1, z2, batch_size=0)

        return loss, None

    def test(self, processed_data):
        data, adj = processed_data
        self.model.eval()
        z = self.model(data, adj)
        kmeans = kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
        preds = kmeans.fit_predict(z).cpu().numpy()
        return preds






# https://github.com/zekarias-tilahun/SelfGNN
import torch
import torch.nn as nn
import torch.nn.functional as F
from fast_pytorch_kmeans import KMeans
import scipy.sparse as sp
from torch_geometric.nn import GCNConv, GATConv, SAGEConv
from functools import wraps
import copy
import ugle
from ugle.trainer import ugleTrainer


class EMA:
    def __init__(self, beta):
        super().__init__()
        self.beta = beta

    def update_average(self, old, new):
        if old is None:
            return new
        return old * self.beta + (1 - self.beta) * new


def loss_fn(x, y):
    x = F.normalize(x, dim=-1, p=2)
    y = F.normalize(y, dim=-1, p=2)
    return 2 - 2 * (x * y).sum(dim=-1)


def singleton(cache_key):
    def inner_fn(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            instance = getattr(self, cache_key)
            if instance is not None:
                return instance

            instance = fn(self, *args, **kwargs)
            setattr(self, cache_key, instance)
            return instance

        return wrapper

    return inner_fn


def update_moving_average(ema_updater, ma_model, current_model):
    for current_params, ma_params in zip(current_model.parameters(), ma_model.parameters()):
        old_weight, up_weight = ma_params.data, current_params.data
        ma_params.data = ema_updater.update_average(old_weight, up_weight)


def set_requires_grad(model, val):
    for p in model.parameters():
        p.requires_grad = val


class Normalize(nn.Module):
    def __init__(self, dim=None, method="batch"):
        super().__init__()
        method = None if dim is None else method
        if method == "batch":
            self.norm = nn.BatchNorm1d(dim)
        elif method == "layer":
            self.norm = nn.LayerNorm(dim)
        else:  # No norm => identity
            self.norm = lambda x: x

    def forward(self, x):
        return self.norm(x)


class Encoder(nn.Module):

    def __init__(self, args, layers, heads, dropout=None):
        super().__init__()
        rep_dim = layers[-1]
        self.gnn_type = args.gnn_type
        self.dropout = dropout
        self.project = args.prj_head_norm

        self.stacked_gnn = get_encoder(args.gnn_type, layers, heads, args.concat)
        self.encoder_norm = Normalize(dim=rep_dim, method=args.encoder_norm)
        if self.project != "no":
            self.projection_head = nn.Sequential(
                nn.Linear(rep_dim, rep_dim),
                Normalize(dim=rep_dim, method=args.prj_head_norm),
                nn.ReLU(inplace=True), nn.Dropout(dropout))

    def forward(self, x, edge_index, edge_weight=None):
        for i, gnn in enumerate(self.stacked_gnn):
            if self.gnn_type == "gat" or self.gnn_type == "sage":
                x = gnn(x, edge_index)
            else:
                x = gnn(x, edge_index, edge_weight=edge_weight)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.encoder_norm(x)
        return x, (self.projection_head(x) if self.project != "no" else None)


class SelfGNN(nn.Module):

    def __init__(self, args, layers, heads, dropout=0.0, moving_average_decay=0.99):
        super().__init__()
        self.student_encoder = Encoder(args, layers=layers, heads=heads, dropout=dropout)
        self.teacher_encoder = None
        self.teacher_ema_updater = EMA(moving_average_decay)
        rep_dim = layers[-1]
        self.student_predictor = nn.Sequential(
            nn.Linear(rep_dim, rep_dim),
            Normalize(dim=rep_dim, method=args.prd_head_norm),
            nn.ReLU(inplace=True),
            nn.Dropout(dropout))

    @singleton('teacher_encoder')
    def _get_teacher_encoder(self):
        teacher_encoder = copy.deepcopy(self.student_encoder)
        set_requires_grad(teacher_encoder, False)
        return teacher_encoder

    def reset_moving_average(self):
        del self.teacher_encoder
        self.teacher_encoder = None

    def update_moving_average(self):
        assert self.teacher_encoder is not None, 'teacher encoder has not been created yet'
        update_moving_average(self.teacher_ema_updater, self.teacher_encoder, self.student_encoder)

    def encode(self, x, edge_index, edge_weight=None, encoder=None):
        encoder = self.student_encoder if encoder is None else encoder
        encoder.train(self.training)
        return encoder(x, edge_index, edge_weight)

    def forward(self, x1, x2, edge_index_v1, edge_index_v2, edge_weight_v1=None, edge_weight_v2=None):
        """
        Apply student network on both views

        v<x>_rep is the output of the stacked GNN
        v<x>_student is the output of the student projection head, if used, otherwise is just a reference to v<x>_rep
        """
        v1_enc = self.encode(x=x1, edge_index=edge_index_v1, edge_weight=edge_weight_v1)
        v1_rep, v1_student = v1_enc if v1_enc[1] is not None else (v1_enc[0], v1_enc[0])
        v2_enc = self.encode(x=x2, edge_index=edge_index_v2, edge_weight=edge_weight_v2)
        v2_rep, v2_student = v2_enc if v2_enc[1] is not None else (v2_enc[0], v2_enc[0])

        """
        Apply the student predictor both views using the outputs from the previous phase 
        (after the stacked GNN or projection head - if there is one)
        """
        v1_pred = self.student_predictor(v1_student)
        v2_pred = self.student_predictor(v2_student)

        """
        Apply the same procedure on the teacher network as in the student network except the predictor.
        """
        with torch.no_grad():
            teacher_encoder = self._get_teacher_encoder()
            v1_enc = self.encode(x=x1, edge_index=edge_index_v1, edge_weight=edge_weight_v1, encoder=teacher_encoder)
            v1_teacher = v1_enc[1] if v1_enc[1] is not None else v1_enc[0]
            v2_enc = self.encode(x=x2, edge_index=edge_index_v2, edge_weight=edge_weight_v2, encoder=teacher_encoder)
            v2_teacher = v2_enc[1] if v2_enc[1] is not None else v2_enc[0]

        """
        Compute symmetric loss (once based on view1 (v1) as input to the student and then using view2 (v2))
        """
        loss1 = loss_fn(v1_pred, v2_teacher.detach())
        loss2 = loss_fn(v2_pred, v1_teacher.detach())

        loss = loss1 + loss2
        return v1_rep, v2_rep, loss.mean()


def get_encoder(gnn_type, layers, heads, concat):
    """
    Builds the GNN backbone as required
    """
    if gnn_type == "gcn":
        return nn.ModuleList([GCNConv(layers[i - 1], layers[i]) for i in range(1, len(layers))])
    elif gnn_type == "sage":
        return nn.ModuleList([SAGEConv(layers[i - 1], layers[i]) for i in range(1, len(layers))])
    elif gnn_type == "gat":
        return nn.ModuleList(
            [GATConv(layers[i - 1], layers[i] // heads[i - 1], heads=heads[i - 1], concat=concat)
             for i in range(1, len(layers))])


class selfgnn_trainer(ugleTrainer):
    def preprocess_data(self, features, adjacency):

        adjacency = sp.csr_matrix(adjacency)

        augmentation = ugle.datasets.Augmentations(method=self.cfg.args.aug)
        features, adjacency, aug_features, aug_adjacency = augmentation(features, adjacency)

        features = torch.FloatTensor(features)
        adjacency = torch.LongTensor(adjacency)
        aug_features = torch.FloatTensor(aug_features)
        aug_adjacency = torch.LongTensor(aug_adjacency)

        diff = abs(aug_features.shape[1] - features.shape[1])
        if diff > 0:
            """
            Data augmentation on the features could lead to mismatch between the shape of the two views,
            hence the smaller view should be padded with zero. (smaller_data is a reference, changes will
            reflect on the original data)
            """
            which_small = 'features' if features.shape[1] < aug_features.shape[1] else 'aug_features'
            smaller_data = features if which_small == 'features' else aug_features
            smaller_data = F.pad(smaller_data, pad=(0, diff))
            if which_small == 'features':
                features = smaller_data
            else:
                aug_features = smaller_data
            features = F.normalize(features)
            aug_features = F.normalize(aug_features)
            self.cfg.args.n_features = features.shape[1]

        return features, adjacency, aug_features, aug_adjacency

    def training_preprocessing(self, args, processed_data):

        layers = [args.n_features, args.layer1, args.layer2]
        heads = [args.head1, args.head2]
        self.model = SelfGNN(args=args, layers=layers, heads=heads).to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)
        self.optimizers = [optimizer]

        return

    def training_epoch_iter(self, args, processed_data):
        features, adjacency, aug_features, aug_adjacency = processed_data
        v1_output, v2_output, loss = self.model(
            x1=features, x2=aug_features, edge_index_v1=adjacency, edge_index_v2=aug_adjacency)
        self.model.update_moving_average()
        return loss, None

    def test(self, processed_data):
        features, adjacency, aug_features, aug_adjacency = processed_data

        self.model.eval()

        v1_output, v2_output, _ = self.model(
            x1=features, x2=aug_features, edge_index_v1=adjacency,
            edge_index_v2=aug_adjacency)

        emb = torch.cat([v1_output, v2_output], dim=1).detach()
        emb = emb.squeeze(0)
        kmeans = kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
        preds = kmeans.fit_predict(emb).cpu().numpy()

        return preds




# https://github.com/Namkyeong/BGRL_Pytorch
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from fast_pytorch_kmeans import KMeans
import scipy.sparse as sp
from torch_geometric.nn import GCNConv
import copy
import warnings
warnings.filterwarnings('ignore')
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=DeprecationWarning)
    from torch_geometric.utils import dropout_adj
import ugle
from ugle.trainer import ugleTrainer

class EMA:
    def __init__(self, beta, epochs):
        super().__init__()
        self.beta = beta
        self.step = 0
        self.total_steps = epochs

    def update_average(self, old, new):
        if old is None:
            return new
        beta = 1 - (1 - self.beta) * (np.cos(np.pi * self.step / self.total_steps) + 1) / 2.0
        self.step += 1
        return old * beta + (1 - beta) * new


def loss_fn(x, y):
    x = F.normalize(x, dim=-1, p=2)
    y = F.normalize(y, dim=-1, p=2)
    return 2 - 2 * (x * y).sum(dim=-1)


def update_moving_average(ema_updater, ma_model, current_model):
    for current_params, ma_params in zip(current_model.parameters(), ma_model.parameters()):
        old_weight, up_weight = ma_params.data, current_params.data
        ma_params.data = ema_updater.update_average(old_weight, up_weight)


def set_requires_grad(model, val):
    for p in model.parameters():
        p.requires_grad = val


class Encoder(nn.Module):

    def __init__(self, layer_config, dropout=None, project=False, **kwargs):
        super().__init__()

        self.conv1 = GCNConv(layer_config[0], layer_config[1])
        self.bn1 = nn.BatchNorm1d(layer_config[1], momentum=0.01)
        self.prelu1 = nn.PReLU()
        self.conv2 = GCNConv(layer_config[1], layer_config[2])
        self.bn2 = nn.BatchNorm1d(layer_config[2], momentum=0.01)
        self.prelu2 = nn.PReLU()

    def forward(self, x, edge_index, edge_weight=None):
        x = self.conv1(x, edge_index, edge_weight=edge_weight)
        x = self.prelu1(self.bn1(x))
        x = self.conv2(x, edge_index, edge_weight=edge_weight)
        x = self.prelu2(self.bn2(x))

        return x


def init_weights(m):
    if type(m) == nn.Linear:
        torch.nn.init.xavier_uniform_(m.weight)
        m.bias.data.fill_(0.01)


class BGRL(nn.Module):

    def __init__(self, layer_config, pred_hid, dropout=0.0, moving_average_decay=0.99, epochs=1000, **kwargs):
        super().__init__()
        self.student_encoder = Encoder(layer_config=layer_config, dropout=dropout, **kwargs)
        self.teacher_encoder = copy.deepcopy(self.student_encoder)
        set_requires_grad(self.teacher_encoder, False)
        self.teacher_ema_updater = EMA(moving_average_decay, epochs)
        rep_dim = layer_config[-1]
        self.student_predictor = nn.Sequential(nn.Linear(rep_dim, pred_hid), nn.PReLU(), nn.Linear(pred_hid, rep_dim))
        self.student_predictor.apply(init_weights)

    def reset_moving_average(self):
        del self.teacher_encoder
        self.teacher_encoder = None

    def update_moving_average(self):
        assert self.teacher_encoder is not None, 'teacher encoder has not been created yet'
        update_moving_average(self.teacher_ema_updater, self.teacher_encoder, self.student_encoder)

    def forward(self, x1, x2, edge_index_v1, edge_index_v2, edge_weight_v1=None, edge_weight_v2=None):
        v1_student = self.student_encoder(x=x1, edge_index=edge_index_v1, edge_weight=edge_weight_v1)
        v2_student = self.student_encoder(x=x2, edge_index=edge_index_v2, edge_weight=edge_weight_v2)

        v1_pred = self.student_predictor(v1_student)
        v2_pred = self.student_predictor(v2_student)

        with torch.no_grad():
            v1_teacher = self.teacher_encoder(x=x1, edge_index=edge_index_v1, edge_weight=edge_weight_v1)
            v2_teacher = self.teacher_encoder(x=x2, edge_index=edge_index_v2, edge_weight=edge_weight_v2)

        loss1 = loss_fn(v1_pred, v2_teacher.detach())
        loss2 = loss_fn(v2_pred, v1_teacher.detach())

        loss = loss1 + loss2
        return v1_student, v2_student, loss.mean()


class bgrl_trainer(ugleTrainer):
    def preprocess_data(self, features, adjacency):
        adj_label = sp.coo_matrix(adjacency)
        adj_label = adj_label.todok()

        outwards = [i[0] for i in adj_label.keys()]
        inwards = [i[1] for i in adj_label.keys()]

        adj = torch.from_numpy(np.array([outwards, inwards], dtype=int))

        data = torch.FloatTensor(features)

        return data, adj

    def training_preprocessing(self, args, processed_data):

        layers = [args.n_features, args.layer1, args.layer2]
        self.model = BGRL(layer_config=layers, pred_hid=args.pred_hidden,
                     dropout=args.dropout, epochs=args.max_epoch).to(self.device)

        optimizer = torch.optim.Adam(params=self.model.parameters(), lr=args.learning_rate, weight_decay=1e-5)
        if args.max_epoch == 1000:
            args.max_epoch += 1
        scheduler = lambda epoch: epoch / 1000 if epoch < 1000 else (1 + np.cos((epoch - 1000) * np.pi / (args.max_epoch - 1000))) * 0.5

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=scheduler)

        self.scheduler = scheduler
        self.optimizers = [optimizer]

        return

    def training_epoch_iter(self, args, processed_data):

        data, adj = processed_data
        self.model.update_moving_average()
        adj_1 = dropout_adj(adj, p=args.drop_edge_rate_1, force_undirected=True)[0]
        adj_2 = dropout_adj(adj, p=args.drop_edge_rate_2, force_undirected=True)[0]

        x_1 = ugle.datasets.aug_drop_features(data, drop_percent=args.drop_feature_rate_1)
        x_2 = ugle.datasets.aug_drop_features(data, drop_percent=args.drop_feature_rate_2)

        v1_output, v2_output, loss = self.model(
            x1=x_1, x2=x_2, edge_index_v1=adj_1, edge_index_v2=adj_2,
            edge_weight_v1=None, edge_weight_v2=None)

        return loss, None

    def test(self, processed_data):
        data, adj = processed_data
        self.model.eval()
        v1_output, v2_output, _ = self.model(
                x1=data, x2=data, edge_index_v1=adj, edge_index_v2=adj,
            edge_weight_v1=None,
            edge_weight_v2=None)

        kmeans = kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
        preds = kmeans.fit_predict(v1_output).cpu().numpy()


        return preds


# https://github.com/GRAND-Lab/SUBLIME
import torch
import torch.nn as nn
from torch.nn import Sequential, Linear, ReLU
import torch.nn.functional as F
import numpy as np
import copy
from fast_pytorch_kmeans import KMeans
from ugle.trainer import ugleTrainer
EOS = 1e-10
import types

class GCNConv_dense(nn.Module):
    def __init__(self, input_size, output_size):
        super(GCNConv_dense, self).__init__()
        self.linear = nn.Linear(input_size, output_size)

    def init_para(self):
        self.linear.reset_parameters()

    def forward(self, input, A, sparse=False):
        hidden = self.linear(input)
        output = torch.matmul(A, hidden)
        return output


class GCN(nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, num_layers, dropout, dropout_adj, Adj, sparse):
        super(GCN, self).__init__()
        self.layers = nn.ModuleList()

        self.layers.append(GCNConv_dense(in_channels, hidden_channels))
        for i in range(num_layers - 2):
            self.layers.append(GCNConv_dense(hidden_channels, hidden_channels))
        self.layers.append(GCNConv_dense(hidden_channels, out_channels))

        self.dropout = dropout
        self.dropout_adj_p = dropout_adj
        self.Adj = Adj
        self.Adj.requires_grad = False

        self.dropout_adj = nn.Dropout(p=dropout_adj)

    def forward(self, x):

        Adj = self.dropout_adj(self.Adj)

        for i, conv in enumerate(self.layers[:-1]):
            x = conv(x, Adj)
            x = F.relu(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.layers[-1](x, Adj)
        return x


class GraphEncoder(nn.Module):
    def __init__(self, nlayers, in_dim, hidden_dim, emb_dim, proj_dim, dropout, dropout_adj):

        super(GraphEncoder, self).__init__()
        self.dropout = dropout
        self.dropout_adj_p = dropout_adj

        self.gnn_encoder_layers = nn.ModuleList()

        self.gnn_encoder_layers.append(GCNConv_dense(in_dim, hidden_dim))
        for _ in range(nlayers - 2):
            self.gnn_encoder_layers.append(GCNConv_dense(hidden_dim, hidden_dim))
        self.gnn_encoder_layers.append(GCNConv_dense(hidden_dim, emb_dim))

        self.dropout_adj = nn.Dropout(p=dropout_adj)

        self.proj_head = Sequential(Linear(emb_dim, proj_dim), ReLU(inplace=True),
                                    Linear(proj_dim, proj_dim))

    def forward(self, x, Adj_, branch=None):

        Adj = self.dropout_adj(Adj_)

        for conv in self.gnn_encoder_layers[:-1]:
            x = conv(x, Adj)
            x = F.relu(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.gnn_encoder_layers[-1](x, Adj)
        z = self.proj_head(x)
        return z, x


class GCL(nn.Module):
    def __init__(self, nlayers, in_dim, hidden_dim, emb_dim, proj_dim, dropout, dropout_adj, sparse=False):
        super(GCL, self).__init__()

        self.encoder = GraphEncoder(nlayers, in_dim, hidden_dim, emb_dim, proj_dim, dropout, dropout_adj)

    def forward(self, x, Adj_, branch=None):
        z, embedding = self.encoder(x, Adj_, branch)
        return z, embedding

    @staticmethod
    def calc_loss(x, x_aug, temperature=0.2, sym=True):
        batch_size, _ = x.size()
        x_abs = x.norm(dim=1)
        x_aug_abs = x_aug.norm(dim=1)

        sim_matrix = torch.einsum('ik,jk->ij', x, x_aug) / torch.einsum('i,j->ij', x_abs, x_aug_abs)
        sim_matrix = torch.exp(sim_matrix / temperature)
        pos_sim = sim_matrix[range(batch_size), range(batch_size)]
        if sym:
            loss_0 = pos_sim / (sim_matrix.sum(dim=0) - pos_sim)
            loss_1 = pos_sim / (sim_matrix.sum(dim=1) - pos_sim)

            loss_0 = - torch.log(loss_0).mean()
            loss_1 = - torch.log(loss_1).mean()
            loss = (loss_0 + loss_1) / 2.0
            return loss
        else:
            loss_1 = pos_sim / (sim_matrix.sum(dim=1) - pos_sim)
            loss_1 = - torch.log(loss_1).mean()
            return loss_1


class MLP_learner(nn.Module):
    def __init__(self, nlayers, isize, k, knn_metric, i, sparse, act):
        super(MLP_learner, self).__init__()

        self.layers = nn.ModuleList()
        if nlayers == 1:
            self.layers.append(nn.Linear(isize, isize))
        else:
            self.layers.append(nn.Linear(isize, isize))
            for _ in range(nlayers - 2):
                self.layers.append(nn.Linear(isize, isize))
            self.layers.append(nn.Linear(isize, isize))

        self.input_dim = isize
        self.output_dim = isize
        self.k = k
        self.knn_metric = knn_metric
        self.non_linearity = 'relu'
        self.param_init()
        self.i = i
        self.act = act

    def internal_forward(self, h):
        for i, layer in enumerate(self.layers):
            h = layer(h)
            if i != (len(self.layers) - 1):
                if self.act == "relu":
                    h = F.relu(h)
                elif self.act == "tanh":
                    h = F.tanh(h)
        return h

    def param_init(self):
        for layer in self.layers:
            layer.weight = nn.Parameter(torch.eye(self.input_dim))

def cal_similarity_graph(node_embeddings):
    similarity_graph = torch.mm(node_embeddings, node_embeddings.t())
    return similarity_graph


def apply_non_linearity(tensor, non_linearity, i):
    if non_linearity == 'elu':
        return F.elu(tensor * i - i) + 1
    elif non_linearity == 'relu':
        return F.relu(tensor)
    elif non_linearity == 'none':
        return tensor
    else:
        raise NameError('We dont support the non-linearity yet')


def symmetrize(adj):  # only for non-sparse
    return (adj + adj.T) / 2


def normalize(adj, mode, sparse=False):

    if mode == "sym":
        inv_sqrt_degree = 1. / (torch.sqrt(adj.sum(dim=1, keepdim=False)) + EOS)
        return inv_sqrt_degree[:, None] * adj * inv_sqrt_degree[None, :]
    elif mode == "row":
        inv_degree = 1. / (adj.sum(dim=1, keepdim=False) + EOS)
        return inv_degree[:, None] * adj
    else:
        exit("wrong norm mode")


def split_batch(init_list, batch_size):
    groups = zip(*(iter(init_list),) * batch_size)
    end_list = [list(i) for i in groups]
    count = len(init_list) % batch_size
    end_list.append(init_list[-count:]) if count != 0 else end_list
    return end_list


def get_feat_mask(features, mask_rate):
    feat_node = features.shape[1]
    mask = torch.zeros(features.shape)
    samples = np.random.choice(feat_node, size=int(feat_node * mask_rate), replace=False)
    mask[:, samples] = 1
    return mask, samples


class gclloss_obj:
    def __init__(self, maskfeat_rate_anchor, maskfeat_rate_learner, contrast_batch_size, device, n_nodes):
        super(gclloss_obj, self).__init__()
        self.maskfeat_rate_anchor = maskfeat_rate_anchor
        self.maskfeat_rate_learner = maskfeat_rate_learner
        self.contrast_batch_size = contrast_batch_size
        self.device = device
        self.n_nodes = n_nodes

    def top_k(self, raw_graph, K):
        values, indices = raw_graph.topk(k=int(K), dim=-1)
        assert torch.max(indices) < raw_graph.shape[1]
        mask = torch.zeros(raw_graph.shape).to(self.device)
        mask[torch.arange(raw_graph.shape[0]).view(-1, 1), indices] = 1.

        mask.requires_grad = False
        sparse_graph = raw_graph * mask
        return sparse_graph

    def graph_learner_forward(self, graph_learner, features):
        embeddings = graph_learner.internal_forward(features)
        embeddings = F.normalize(embeddings, dim=1, p=2)
        similarities = cal_similarity_graph(embeddings)
        similarities = self.top_k(similarities, graph_learner.k + 1)
        similarities = apply_non_linearity(similarities, graph_learner.non_linearity, graph_learner.i)
        return similarities

    def loss_gcl(self, model, graph_learner, features, anchor_adj):

        # view 1: anchor graph
        if self.maskfeat_rate_anchor:
            mask_v1, _ = get_feat_mask(features, self.maskfeat_rate_anchor)
            mask_v1 = mask_v1.to(self.device)
            features_v1 = features * (1 - mask_v1)
        else:
            features_v1 = features.copy()

        features_v1 = features_v1.to(self.device)
        z1, _ = model(features_v1, anchor_adj, 'anchor')

        # view 2: learned graph
        if self.maskfeat_rate_learner:
            mask, _ = get_feat_mask(features, self.maskfeat_rate_learner)
            mask = mask.to(self.device)
            features_v2 = features * (1 - mask)
        else:
            features_v2 = features.copy()

        learned_adj = self.graph_learner_forward(graph_learner, features)


        learned_adj = symmetrize(learned_adj)
        learned_adj = normalize(learned_adj, 'sym')

        z2, _ = model(features_v2, learned_adj, 'learner')

        # compute loss
        if self.contrast_batch_size:
            node_idxs = list(range(self.n_nodes))
            # random.shuffle(node_idxs)
            batches = split_batch(node_idxs, self.contrast_batch_size)
            loss = 0
            for batch in batches:
                weight = len(batch) / self.n_nodes
                loss += model.calc_loss(z1[batch], z2[batch]) * weight
        else:
            loss = model.calc_loss(z1, z2)

        return loss, learned_adj


    

class sublime_trainer(ugleTrainer):
    def preprocess_data(self, features, adjacency):

        anchor_adj_raw = torch.from_numpy(adjacency)
        anchor_adj = normalize(anchor_adj_raw, 'sym')
        features = torch.FloatTensor(features)
        anchor_adj = anchor_adj.float()

        return features, anchor_adj

    def training_preprocessing(self, args, processed_data):
        if args.learner_type == 'mlp':
            self.graph_learner = MLP_learner(args.nlayers, args.n_features, args.k, args.sim_function, args.i, args.sparse,
                                        args.activation_learner).to(self.device)

        self.model = GCL(nlayers=args.nlayers, in_dim=args.n_features, hidden_dim=args.hidden_dim,
                    emb_dim=args.rep_dim, proj_dim=args.proj_dim,
                    dropout=args.dropout, dropout_adj=args.dropedge_rate).to(self.device)

        optimizer_cl = torch.optim.Adam(self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)
        optimizer_learner = torch.optim.Adam(self.graph_learner.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)
        self.optimizers = [optimizer_cl, optimizer_learner]
        self.gcl_obj = gclloss_obj(self.cfg.args.maskfeat_rate_anchor, 
                                   self.cfg.args.maskfeat_rate_learner, 
                                   self.cfg.args.contrast_batch_size,
                                   self.device,
                                   self.cfg.args.n_nodes)

        return

    def training_epoch_iter(self, args, processed_data):
        features, anchor_adj = processed_data

        loss, Adj = self.gcl_obj.loss_gcl(self.model, self.graph_learner, features, anchor_adj)

        # Structure Bootstrapping
        if (1 - args.tau) and (args.c == 0 or self.current_epoch % args.c == 0):
            anchor_adj = anchor_adj * args.tau + Adj.detach() * (1 - args.tau)

        processed_data = (features, anchor_adj)

        return loss, processed_data

    def test(self, processed_data):
        features, anchor_adj = processed_data

        self.model.eval()
        self.graph_learner.eval()
        with torch.no_grad():
            _, Adj = self.gcl_obj.loss_gcl(self.model, self.graph_learner, features, anchor_adj)
            _, embedding = self.model(features, Adj)
            embedding = embedding.squeeze(0)

            kmeans = kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
            preds = kmeans.fit_predict(embedding).cpu().numpy()

        return preds



from setuptools import setup, find_packages

PREREQS = [ "torch==1.12.0" ]

from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info

def requires( packages ): 
    from os import system
    from sys import executable as PYTHON_PATH
    from pkg_resources import require
    require( "pip" )
    CMD_TMPLT = '"' + PYTHON_PATH + '" -m pip install %s'
    for pkg in packages: system( CMD_TMPLT % (pkg,) )       

class OrderedInstall( install ):
    def run( self ):
        requires( PREREQS )
        install.run( self )        

class OrderedDevelop( develop ):
    def run( self ):
        requires( PREREQS )
        develop.run( self )        

class OrderedEggInfo( egg_info ):
    def run( self ):
        requires( PREREQS )
        egg_info.run( self )        

CMD_CLASSES = { 
     "install" : OrderedInstall
   , "develop" : OrderedDevelop
   , "egg_info": OrderedEggInfo 
}        

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory /  ".github/README.md").read_text()

setup(name='ugle', 
      description='This is a repository for investigating implementations of GNNs for \
                   unsupervised clustering.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/willleeney/ugle',
      author='William Leeney',
      author_email='will.leeney@outlook.com',
      license_files = ('LICENSE',),
      version='0.6.0', 
      packages=find_packages(),
      install_requires=['pytest',
                        'omegaconf',
                        'torch==1.12.0',
                        'torch_geometric',
                        'torch_scatter',
                        'torch_sparse',
                        'torch_cluster',
                        'torch_spline_conv',
                        'numpy<=1.23.0',
                        'networkx',
                        'gdown',
                        'optuna',
                        'memory_profiler',
                        'line_profiler',
                        'matplotlib',
                        'fast-pytorch-kmeans',
                        'names_generator'],
      cmdclass = CMD_CLASSES,
      python_requires=">=3.9.12",
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
       ' Intended Audience :: Information Technology', 
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.9',
    ],

)



# https://github.com/kavehhassani/mvgrl
import torch
import torch.nn as nn
import ugle
import scipy.sparse as sp
import numpy as np
from fast_pytorch_kmeans import KMeans
from sklearn.preprocessing import MinMaxScaler
from ugle.trainer import ugleTrainer
from ugle.gnn_architecture import GCN, AvgReadout, mvgrl_Discriminator
from ugle.process import sparse_mx_to_torch_sparse_tensor


class Model(nn.Module):
    def __init__(self, n_in, n_h, act):
        super(Model, self).__init__()
        self.gcn1 = GCN(n_in, n_h, act)
        self.gcn2 = GCN(n_in, n_h, act)
        self.read = AvgReadout()

        self.sigm = nn.Sigmoid()

        self.disc = mvgrl_Discriminator(n_h)

    def forward(self, seq1, seq2, adj, diff, sparse, msk, samp_bias1, samp_bias2):
        h_1 = self.gcn1(seq1, adj, sparse)
        c_1 = self.read(h_1, msk)
        c_1 = self.sigm(c_1)

        h_2 = self.gcn2(seq1, diff, sparse)
        c_2 = self.read(h_2, msk)
        c_2 = self.sigm(c_2)

        h_3 = self.gcn1(seq2, adj, sparse)
        h_4 = self.gcn2(seq2, diff, sparse)

        ret = self.disc(c_1, c_2, h_1, h_2, h_3, h_4, samp_bias1, samp_bias2)

        return ret, h_1, h_2

    def embed(self, seq, adj, diff, sparse, msk):
        h_1 = self.gcn1(seq, adj, sparse)
        c = self.read(h_1, msk)

        h_2 = self.gcn2(seq, diff, sparse)
        return (h_1 + h_2).detach(), c.detach()


class mvgrl_trainer(ugleTrainer):
    def preprocess_data(self, features, adjacency):
        epsilons = [1e-5, 1e-4, 1e-3, 1e-2]

        diff_adj = ugle.process.compute_ppr(adjacency)
        avg_degree = np.sum(adjacency) / adjacency.shape[0]
        epsilon = epsilons[np.argmin([abs(avg_degree - np.argwhere(diff_adj >= e).shape[0] / diff_adj.shape[0])
                                      for e in epsilons])]

        diff_adj[diff_adj < epsilon] = 0.0
        scaler = MinMaxScaler()
        scaler.fit(diff_adj)
        diff_adj = scaler.transform(diff_adj)

        features = ugle.process.preprocess_features(features)
        adjacency = ugle.process.normalize_adj(adjacency + sp.eye(adjacency.shape[0])).toarray()

        return features, adjacency, diff_adj

    def training_preprocessing(self, args, processed_data):
        features, adj, diff_adj = processed_data

        if adj.shape[-1] < args.sample_size:
            args.sample_size = int(np.floor(adj.shape[-1] / 100.0) * 100)

        self.model = Model(args.n_features, args.hid_units, args.activation).to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=args.learning_rate, weight_decay=args.weight_decay)
        self.optimizers = [optimizer]
        self.loss_function = nn.BCEWithLogitsLoss()

        return

    def training_epoch_iter(self, args, processed_data):

        features, adj, diff_adj = processed_data

        if adj.shape[-1] < self.cfg.args.sample_size:
            self.cfg.args.sample_size = int(np.floor(adj.shape[-1] / 100.0) * 100)

        lbl_1 = torch.ones(self.cfg.args.batch_size, self.cfg.args.sample_size * 2)
        lbl_2 = torch.zeros(self.cfg.args.batch_size, self.cfg.args.sample_size * 2)
        lbl = torch.cat((lbl_1, lbl_2), 1).to(self.device)
        idx = np.random.randint(0, adj.shape[-1] - args.sample_size + 1, args.batch_size)

        ba, bd, bf = [], [], []
        for i in idx:
            ba.append(adj[i: i + args.sample_size, i: i + args.sample_size])
            bd.append(diff_adj[i: i + args.sample_size, i: i + args.sample_size])
            bf.append(features[i: i + args.sample_size])

        ba = np.asarray(ba).reshape(args.batch_size, args.sample_size, args.sample_size)
        bd = np.array(bd).reshape(args.batch_size, args.sample_size, args.sample_size)
        bf = np.array(bf).reshape(args.batch_size, args.sample_size, args.n_features)

        if args.sparse:
            ba = sparse_mx_to_torch_sparse_tensor(sp.coo_matrix(ba))
            bd = sparse_mx_to_torch_sparse_tensor(sp.coo_matrix(bd))
        else:
            ba = torch.FloatTensor(ba)
            bd = torch.FloatTensor(bd)

        bf = torch.FloatTensor(bf)
        idx = np.random.permutation(args.sample_size)
        shuf_fts = bf[:, idx, :].to(self.device)
        ba = ba.to(self.device)
        bd = bd.to(self.device)
        bf = bf.to(self.device)

        logits, _, _ = self.model(bf, shuf_fts, ba, bd, args.sparse, None, None, None)

        loss = self.loss_function(logits, lbl)

        return loss, None

    def test(self, processed_data):
        features, adj, diff_adj = processed_data

        if self.cfg.args.sparse:
            adj = sparse_mx_to_torch_sparse_tensor(sp.coo_matrix(adj))
            diff_adj = sparse_mx_to_torch_sparse_tensor(sp.coo_matrix(diff_adj))

        features = torch.FloatTensor(features[np.newaxis]).to(self.device)
        adj = torch.FloatTensor(adj[np.newaxis]).to(self.device)
        diff_adj = torch.FloatTensor(diff_adj[np.newaxis]).to(self.device)

        embeds, _ = self.model.embed(features, adj, diff_adj, self.cfg.args.sparse, None)
        embeds = embeds.squeeze(0)

        kmeans = kmeans = KMeans(n_clusters=self.cfg.args.n_clusters)
        preds = kmeans.fit_predict(embeds).cpu().numpy()

        return preds



from omegaconf import OmegaConf, open_dict, DictConfig
from optuna import Study
from typing import Tuple
import random
import torch
import numpy as np
from ugle.logger import log, ugle_path
import optuna
from optuna.trial import Trial
import os 
import pickle

neural_algorithms = ['daegc', 'dgi', 'dmon', 'grace', 'mvgrl', 'selfgnn', 'sublime', 'bgrl', 'vgaer', 'cagc']


def load_model_config(override_model: str = None, override_cfg: DictConfig = None) -> DictConfig:
    """
    Loads the base model config, model-specific config, followed by override_cfg.

    Args:
        override_model (str): name of the model to load config for
        override_cfg (DictConfig): Config to override options on the loaded config

    Returns:
        config (DictConfig): Config for running experiments
    """ 

    config = OmegaConf.load(f"{ugle_path}/ugle/configs/config.yaml")

    if override_model:
        config.model = override_model

    config_name = config.model
    model_name = config.model
    
    # model_path structure edit if testing
    if config_name.__contains__('_'):
        model_name, name_ext = config_name.split("_")
        config.model = model_name
        model_path = f"{ugle_path}/ugle/configs/models/{model_name}/{config_name}.yaml"
    else:
        model_path = f"{ugle_path}/ugle/configs/models/{model_name}/{model_name}.yaml"

    config = merge_yaml(config, model_path)

    # model_eval override
    if override_cfg:
        with open_dict(config):
            config = OmegaConf.merge(config, override_cfg)

    return config


def process_study_cfg_parameters(config: DictConfig) -> DictConfig:
    """
    Creates optimization directions for study.

    Args:
        config (DictConfig): Config object

    Returns:
        config (DictConfig): Processed config
    """
    opt_directions = []
    if len(config.trainer.valid_metrics) > 1:
        config.trainer.multi_objective_study = True
    else:
        config.trainer.multi_objective_study = False

    for metric in config.trainer.valid_metrics:
        if metric == 'nmi' or metric == 'f1' or metric == 'modularity':
            opt_directions.append('maximize')
        else:
            opt_directions.append('minimize')

    config.trainer.optimisation_directions = opt_directions
    return config


def extract_best_trials_info(study: Study, valid_metrics: list) -> Tuple[list, list]:
    """
    Extracts the best trial from a study for each given metric and associated trial.

    Args:
        study (Study): The study object
        valid_metrics (list): The validation metrics 

    Returns:
        best_values (list): The best values for each metric
        associated_trial (list): The associated trial with each best value
    """
    best_values = study.best_trials[0].values
    associated_trial = [study.best_trials[0].number] * len(valid_metrics)
    if len(study.best_trials) > 1:
        for a_best_trial in study.best_trials[1:]:
            for i, bval in enumerate(a_best_trial.values):
                if (bval > best_values[i] and study.directions[i].name == 'MAXIMIZE') or (
                        bval < best_values[i] and study.directions[i].name == 'MINIMIZE'):
                    best_values[i] = bval
                    associated_trial[i] = a_best_trial.number
                
    return best_values, associated_trial


def merge_yaml(config: DictConfig, yaml_str: str) -> DictConfig:
    """
    Merges config object with config specified in YAML path string.

    Args:
        config (DictConfig): Config object to be merged
        yaml_str (str): Path location of .yaml string to be merged

    Returns:
        config (DictConfig): Merged config
    """
    yaml_dict = OmegaConf.load(yaml_str)
    with open_dict(config):
        config = OmegaConf.merge(config, yaml_dict)
    return config


def set_random(random_seed: int):
    """
    Sets the random seed.
    
    Args:
        random_seed (int): Random seed to be set
    """
    random.seed(random_seed)
    np.random.seed(random_seed)
    torch.manual_seed(random_seed)
    torch.cuda.manual_seed_all(random_seed)
    return


def set_device(gpu: int):
    """
    Returns the correct device.

    Args:
        gpu (int): GPU index to specify the device
    """
    if gpu != -1 and torch.cuda.is_available():
        device = f'cuda:{gpu}'
    else:
        device = 'cpu'
    return device


def remove_last_line():
    """
    Removes the last line in the terminal.

    Note: This function prints ANSI escape sequences to move the cursor to the beginning of the last line.
    """
    print("\033[2F\033[", flush=True)
    return


def sample_hyperparameters(trial: optuna.trial.Trial, args: DictConfig, prune_params=None) -> DictConfig:
    """
    Iterates through the args configuration. If an item is a list, suggests a value based on
    the optuna trial instance.

    Args:
        trial (optuna.trial.Trial): Instance of trial for suggesting parameters
        args (DictConfig): Config dictionary where list
        prune_params: (optional) Parameters to prune from the sampled configuration

    Returns:
        DictConfig: New config with values replaced where list is given
    """
    vars_to_set = []
    vals_to_set = []
    for k, v in args.items_ex(resolve=False):
        if not OmegaConf.is_list(v):
            continue
        else:
            trial_sample = trial.suggest_categorical(k, OmegaConf.to_object(v))
            setattr(args, k, trial_sample)
            vars_to_set.append(k)
            vals_to_set.append(trial_sample)

    if prune_params:
        repeated = prune_params.check_params()
    for var, val in zip(vars_to_set, vals_to_set):
        log.info(f"args.{var}={val}")

    return args

def assign_test_params(config: DictConfig, best_params: dict) -> DictConfig:
    """
    Assigns the best parameters from the hyperparameter selection and assigns test config settings.

    Args:
        config (DictConfig): Original config for training
        best_params (dict): The best hyperparameters from training

    Returns:
        DictConfig: Config for training and then testing
    """
    cfg = config.copy()
    # assigns the best parameters from hp to config
    for key, value in cfg.args.items_ex(resolve=False):
        if not OmegaConf.is_list(value):
            continue
        else:
            cfg.args[key] = best_params[key]

    # assigns parameters for testing
    cfg.trainer.train_to_valid_split = 1.0
    cfg.trainer.n_valid_splits = 1
    cfg.trainer.currently_testing = True

    return cfg


def is_neural(model_name: str) -> bool:
    """
    Checks if an algorithm is neural or non-neural.

    Args:
        model_name (str): The model name string

    Returns:
        bool: True if neural, False if not
    """
    if model_name.__contains__('_'):
        adjusted_model_name, _ = model_name.split("_")
    else:
        adjusted_model_name = 'this is definitely not a name of an algorithm'

    if model_name in neural_algorithms or adjusted_model_name in neural_algorithms:
        return True
    else:
        raise ValueError(f"Algorithm {model_name} is not implemented")


def collate_study_results(average_results: dict, results: dict, val_idx: int) -> dict:
    """
    Converts study results from results.

    Args:
        average_results (dict): Dictionary of metrics with numpy arrays of results across training runs
        results (dict): Dictionary of results
        val_idx (int): Number of the training run

    Returns:
        average_results (dict): Dictionary of metrics with numpy arrays of results across training runs
    """
    for trial in results:
        for metric, value in trial["results"].items():
            average_results[metric][val_idx] = value

    return average_results


def create_study_tracker(k_splits: int, metrics: list) -> dict:
    """
    Creates a study tracker for multiple seeds.

    Args:
        k_splits (int): The number of seeds
        metrics (list): The metrics to create a tracker for

    Returns:
        results (object): The results tracker
    """
    results = {}
    for metric in metrics:
        results[metric] = np.zeros(k_splits)
    return results


def create_experiment_tracker(exp_cfg: DictConfig) -> list:
    """
    Creates the experiment tracker to track results.

    Args:
        exp_cfg (DictConfig): Experiment config

    Returns:
        experiment_tracker (list): List of results objects that store results from individual experiments
    """
    experiment_tracker = []
    if exp_cfg.dataset_algo_combinations:
        for dataset_algo in exp_cfg.dataset_algo_combinations:
            dataset, algorithm = dataset_algo.split('_', 1)
            experiment_tracker.append(OmegaConf.create(
                                {'dataset': dataset,
                                'algorithm': algorithm,
                                'seeds': exp_cfg.seeds,
                                'results': {}
                                }))
            
    elif exp_cfg.datasets and exp_cfg.algorithms:
        for dataset in exp_cfg.datasets:
            for algorithm in exp_cfg.algorithms:
                experiment_tracker.append(OmegaConf.create(
                                {'dataset': dataset,
                                'algorithm': algorithm,
                                'seeds': exp_cfg.seeds,
                                'results': {}
                                }))

    return experiment_tracker



def extract_result_values(exp_result, metrics=['f1', 'nmi', 'modularity', 'conductance']) -> str:
    """
    Extracts the results into a string ready to print for LaTeX display. If std values exist, then they are also used.

    Args:
        exp_result (dict): Dictionary of results
        metrics (list): List of metrics in results (default: ['f1', 'nmi', 'modularity', 'conductance'])

    Returns:
        metric_string (str): String to print given results
    """
    metric_string = ''
    for metric in metrics:
        metric_string += ' & '
        metric_mean = f'{metric}_mean'
        metric_val = exp_result[metric_mean]
        metric_string += str(metric_val)

        metric_std = f'{metric}_std'
        if metric_std in exp_result.keys():
            std = str(exp_result[metric_std])
            metric_string += f'\pm {std}'

    metric_string += ' \\\\'
    return metric_string


def save_experiment_tracker(result_tracker: list, results_path: str):
    """
    Pickle saves result tracker to results path.

    Args:
        result_tracker: List of experiment results
        results_path: The path to save results in
    """
    if not os.path.exists(results_path):
        os.makedirs(results_path)
    experiment_identifier = log.name
    log.info(f'saving to {results_path}{experiment_identifier}')
    pickle.dump(result_tracker, open(f"{results_path}{experiment_identifier}.pkl", "wb"))

    return


def calc_average_results(results_tracker: dict) -> dict:
    """
    Calculates the average of the performance statistic to an appropriate significant figure.

    Args:
        results_tracker (dict): Contains results over seeds

    Returns:
        average_results (dict): The average of results over all seeds
    """
    average_results = {}
    for stat, values in results_tracker.items():
        if stat != 'memory':
            average_results[f'{stat}_mean'] = float(np.format_float_positional(np.mean(values, axis=0),
                                                                               precision=4,
                                                                               unique=False,
                                                                               fractional=False))
            if len(values) > 1:
                average_results[f'{stat}_std'] = float(np.format_float_positional(np.std(values, axis=0),
                                                                                  precision=2,
                                                                                  unique=False,
                                                                                  fractional=False))
        else:
            average_results[f'memory_max'] = float(np.format_float_positional(np.max(values, axis=0),
                                                                              precision=7,
                                                                              unique=False,
                                                                              fractional=False))
    return average_results